# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Interactions with the Juju environment"
# Copyright 2013 Canonical Ltd.
#
# Authors:
#  Charm Helpers Developers <juju@lists.ubuntu.com>

from __future__ import print_function
import copy
from distutils.version import LooseVersion
from enum import Enum
from functools import wraps
from collections import namedtuple
import glob
import os
import json
import yaml
import re
import subprocess
import sys
import errno
import tempfile
from subprocess import CalledProcessError

from charmhelpers import deprecate

import six
if not six.PY3:
    from UserDict import UserDict
else:
    from collections import UserDict


CRITICAL = "CRITICAL"
ERROR = "ERROR"
WARNING = "WARNING"
INFO = "INFO"
DEBUG = "DEBUG"
TRACE = "TRACE"
MARKER = object()
SH_MAX_ARG = 131071


RANGE_WARNING = ('Passing NO_PROXY string that includes a cidr. '
                 'This may not be compatible with software you are '
                 'running in your shell.')


class WORKLOAD_STATES(Enum):
    ACTIVE = 'active'
    BLOCKED = 'blocked'
    MAINTENANCE = 'maintenance'
    WAITING = 'waiting'


cache = {}


def cached(func):
    """Cache return values for multiple executions of func + args

    For example::

        @cached
        def unit_get(attribute):
            pass

        unit_get('test')

    will cache the result of unit_get + 'test' for future calls.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        global cache
        key = json.dumps((func, args, kwargs), sort_keys=True, default=str)
        try:
            return cache[key]
        except KeyError:
            pass  # Drop out of the exception handler scope.
        res = func(*args, **kwargs)
        cache[key] = res
        return res
    wrapper._wrapped = func
    return wrapper


def flush(key):
    """Flushes any entries from function cache where the
    key is found in the function+args """
    flush_list = []
    for item in cache:
        if key in item:
            flush_list.append(item)
    for item in flush_list:
        del cache[item]


def log(message, level=None):
    """Write a message to the juju log"""
    command = ['juju-log']
    if level:
        command += ['-l', level]
    if not isinstance(message, six.string_types):
        message = repr(message)
    command += [message[:SH_MAX_ARG]]
    # Missing juju-log should not cause failures in unit tests
    # Send log output to stderr
    try:
        subprocess.call(command)
    except OSError as e:
        if e.errno == errno.ENOENT:
            if level:
                message = "{}: {}".format(level, message)
            message = "juju-log: {}".format(message)
            print(message, file=sys.stderr)
        else:
            raise


def function_log(message):
    """Write a function progress message"""
    command = ['function-log']
    if not isinstance(message, six.string_types):
        message = repr(message)
    command += [message[:SH_MAX_ARG]]
    # Missing function-log should not cause failures in unit tests
    # Send function_log output to stderr
    try:
        subprocess.call(command)
    except OSError as e:
        if e.errno == errno.ENOENT:
            message = "function-log: {}".format(message)
            print(message, file=sys.stderr)
        else:
            raise


class Serializable(UserDict):
    """Wrapper, an object that can be serialized to yaml or json"""

    def __init__(self, obj):
        # wrap the object
        UserDict.__init__(self)
        self.data = obj

    def __getattr__(self, attr):
        # See if this object has attribute.
        if attr in ("json", "yaml", "data"):
            return self.__dict__[attr]
        # Check for attribute in wrapped object.
        got = getattr(self.data, attr, MARKER)
        if got is not MARKER:
            return got
        # Proxy to the wrapped object via dict interface.
        try:
            return self.data[attr]
        except KeyError:
            raise AttributeError(attr)

    def __getstate__(self):
        # Pickle as a standard dictionary.
        return self.data

    def __setstate__(self, state):
        # Unpickle into our wrapper.
        self.data = state

    def json(self):
        """Serialize the object to json"""
        return json.dumps(self.data)

    def yaml(self):
        """Serialize the object to yaml"""
        return yaml.dump(self.data)


def execution_environment():
    """A convenient bundling of the current execution context"""
    context = {}
    context['conf'] = config()
    if relation_id():
        context['reltype'] = relation_type()
        context['relid'] = relation_id()
        context['rel'] = relation_get()
    context['unit'] = local_unit()
    context['rels'] = relations()
    context['env'] = os.environ
    return context


def in_relation_hook():
    """Determine whether we're running in a relation hook"""
    return 'JUJU_RELATION' in os.environ


def relation_type():
    """The scope for the current relation hook"""
    return os.environ.get('JUJU_RELATION', None)


@cached
def relation_id(relation_name=None, service_or_unit=None):
    """The relation ID for the current or a specified relation"""
    if not relation_name and not service_or_unit:
        return os.environ.get('JUJU_RELATION_ID', None)
    elif relation_name and service_or_unit:
        service_name = service_or_unit.split('/')[0]
        for relid in relation_ids(relation_name):
            remote_service = remote_service_name(relid)
            if remote_service == service_name:
                return relid
    else:
        raise ValueError('Must specify neither or both of relation_name and service_or_unit')


def local_unit():
    """Local unit ID"""
    return os.environ['JUJU_UNIT_NAME']


def remote_unit():
    """The remote unit for the current relation hook"""
    return os.environ.get('JUJU_REMOTE_UNIT', None)


def application_name():
    """
    The name of the deployed application this unit belongs to.
    """
    return local_unit().split('/')[0]


def service_name():
    """
    .. deprecated:: 0.19.1
       Alias for :func:`application_name`.
    """
    return application_name()


def model_name():
    """
    Name of the model that this unit is deployed in.
    """
    return os.environ['JUJU_MODEL_NAME']


def model_uuid():
    """
    UUID of the model that this unit is deployed in.
    """
    return os.environ['JUJU_MODEL_UUID']


def principal_unit():
    """Returns the principal unit of this unit, otherwise None"""
    # Juju 2.2 and above provides JUJU_PRINCIPAL_UNIT
    principal_unit = os.environ.get('JUJU_PRINCIPAL_UNIT', None)
    # If it's empty, then this unit is the principal
    if principal_unit == '':
        return os.environ['JUJU_UNIT_NAME']
    elif principal_unit is not None:
        return principal_unit
    # For Juju 2.1 and below, let's try work out the principle unit by
    # the various charms' metadata.yaml.
    for reltype in relation_types():
        for rid in relation_ids(reltype):
            for unit in related_units(rid):
                md = _metadata_unit(unit)
                if not md:
                    continue
                subordinate = md.pop('subordinate', None)
                if not subordinate:
                    return unit
    return None


@cached
def remote_service_name(relid=None):
    """The remote service name for a given relation-id (or the current relation)"""
    if relid is None:
        unit = remote_unit()
    else:
        units = related_units(relid)
        unit = units[0] if units else None
    return unit.split('/')[0] if unit else None


def hook_name():
    """The name of the currently executing hook"""
    return os.environ.get('JUJU_HOOK_NAME', os.path.basename(sys.argv[0]))


class Config(dict):
    """A dictionary representation of the charm's config.yaml, with some
    extra features:

    - See which values in the dictionary have changed since the previous hook.
    - For values that have changed, see what the previous value was.
    - Store arbitrary data for use in a later hook.

    NOTE: Do not instantiate this object directly - instead call
    ``hookenv.config()``, which will return an instance of :class:`Config`.

    Example usage::

        >>> # inside a hook
        >>> from charmhelpers.core import hookenv
        >>> config = hookenv.config()
        >>> config['foo']
        'bar'
        >>> # store a new key/value for later use
        >>> config['mykey'] = 'myval'


        >>> # user runs `juju set mycharm foo=baz`
        >>> # now we're inside subsequent config-changed hook
        >>> config = hookenv.config()
        >>> config['foo']
        'baz'
        >>> # test to see if this val has changed since last hook
        >>> config.changed('foo')
        True
        >>> # what was the previous value?
        >>> config.previous('foo')
        'bar'
        >>> # keys/values that we add are preserved across hooks
        >>> config['mykey']
        'myval'

    """
    CONFIG_FILE_NAME = '.juju-persistent-config'

    def __init__(self, *args, **kw):
        super(Config, self).__init__(*args, **kw)
        self.implicit_save = True
        self._prev_dict = None
        self.path = os.path.join(charm_dir(), Config.CONFIG_FILE_NAME)
        if os.path.exists(self.path) and os.stat(self.path).st_size:
            self.load_previous()
        atexit(self._implicit_save)

    def load_previous(self, path=None):
        """Load previous copy of config from disk.

        In normal usage you don't need to call this method directly - it
        is called automatically at object initialization.

        :param path:

            File path from which to load the previous config. If `None`,
            config is loaded from the default location. If `path` is
            specified, subsequent `save()` calls will write to the same
            path.

        """
        self.path = path or self.path
        with open(self.path) as f:
            try:
                self._prev_dict = json.load(f)
            except ValueError as e:
                log('Unable to parse previous config data - {}'.format(str(e)),
                    level=ERROR)
        for k, v in copy.deepcopy(self._prev_dict).items():
            if k not in self:
                self[k] = v

    def changed(self, key):
        """Return True if the current value for this key is different from
        the previous value.

        """
        if self._prev_dict is None:
            return True
        return self.previous(key) != self.get(key)

    def previous(self, key):
        """Return previous value for this key, or None if there
        is no previous value.

        """
        if self._prev_dict:
            return self._prev_dict.get(key)
        return None

    def save(self):
        """Save this config to disk.

        If the charm is using the :mod:`Services Framework <services.base>`
        or :meth:'@hook <Hooks.hook>' decorator, this
        is called automatically at the end of successful hook execution.
        Otherwise, it should be called directly by user code.

        To disable automatic saves, set ``implicit_save=False`` on this
        instance.

        """
        with open(self.path, 'w') as f:
            os.fchmod(f.fileno(), 0o600)
            json.dump(self, f)

    def _implicit_save(self):
        if self.implicit_save:
            self.save()


_cache_config = None


def config(scope=None):
    """
    Get the juju charm configuration (scope==None) or individual key,
    (scope=str).  The returned value is a Python data structure loaded as
    JSON from the Juju config command.

    :param scope: If set, return the value for the specified key.
    :type scope: Optional[str]
    :returns: Either the whole config as a Config, or a key from it.
    :rtype: Any
    """
    global _cache_config
    config_cmd_line = ['config-get', '--all', '--format=json']
    try:
        # JSON Decode Exception for Python3.5+
        exc_json = json.decoder.JSONDecodeError
    except AttributeError:
        # JSON Decode Exception for Python2.7 through Python3.4
        exc_json = ValueError
    try:
        if _cache_config is None:
            config_data = json.loads(
                subprocess.check_output(config_cmd_line).decode('UTF-8'))
            _cache_config = Config(config_data)
        if scope is not None:
            return _cache_config.get(scope)
        return _cache_config
    except (exc_json, UnicodeDecodeError) as e:
        log('Unable to parse output from config-get: config_cmd_line="{}" '
            'message="{}"'
            .format(config_cmd_line, str(e)), level=ERROR)
        return None


@cached
def relation_get(attribute=None, unit=None, rid=None):
    """Get relation information"""
    _args = ['relation-get', '--format=json']
    if rid:
        _args.append('-r')
        _args.append(rid)
    _args.append(attribute or '-')
    if unit:
        _args.append(unit)
    try:
        return json.loads(subprocess.check_output(_args).decode('UTF-8'))
    except ValueError:
        return None
    except CalledProcessError as e:
        if e.returncode == 2:
            return None
        raise


def relation_set(relation_id=None, relation_settings=None, **kwargs):
    """Set relation information for the current unit"""
    relation_settings = relation_settings if relation_settings else {}
    relation_cmd_line = ['relation-set']
    accepts_file = "--file" in subprocess.check_output(
        relation_cmd_line + ["--help"], universal_newlines=True)
    if relation_id is not None:
        relation_cmd_line.extend(('-r', relation_id))
    settings = relation_settings.copy()
    settings.update(kwargs)
    for key, value in settings.items():
        # Force value to be a string: it always should, but some call
        # sites pass in things like dicts or numbers.
        if value is not None:
            settings[key] = "{}".format(value)
    if accepts_file:
        # --file was introduced in Juju 1.23.2. Use it by default if
        # available, since otherwise we'll break if the relation data is
        # too big. Ideally we should tell relation-set to read the data from
        # stdin, but that feature is broken in 1.23.2: Bug #1454678.
        with tempfile.NamedTemporaryFile(delete=False) as settings_file:
            settings_file.write(yaml.safe_dump(settings).encode("utf-8"))
        subprocess.check_call(
            relation_cmd_line + ["--file", settings_file.name])
        os.remove(settings_file.name)
    else:
        for key, value in settings.items():
            if value is None:
                relation_cmd_line.append('{}='.format(key))
            else:
                relation_cmd_line.append('{}={}'.format(key, value))
        subprocess.check_call(relation_cmd_line)
    # Flush cache of any relation-gets for local unit
    flush(local_unit())


def relation_clear(r_id=None):
    ''' Clears any relation data already set on relation r_id '''
    settings = relation_get(rid=r_id,
                            unit=local_unit())
    for setting in settings:
        if setting not in ['public-address', 'private-address']:
            settings[setting] = None
    relation_set(relation_id=r_id,
                 **settings)


@cached
def relation_ids(reltype=None):
    """A list of relation_ids"""
    reltype = reltype or relation_type()
    relid_cmd_line = ['relation-ids', '--format=json']
    if reltype is not None:
        relid_cmd_line.append(reltype)
        return json.loads(
            subprocess.check_output(relid_cmd_line).decode('UTF-8')) or []
    return []


@cached
def related_units(relid=None):
    """A list of related units"""
    relid = relid or relation_id()
    units_cmd_line = ['relation-list', '--format=json']
    if relid is not None:
        units_cmd_line.extend(('-r', relid))
    return json.loads(
        subprocess.check_output(units_cmd_line).decode('UTF-8')) or []


def expected_peer_units():
    """Get a generator for units we expect to join peer relation based on
    goal-state.

    The local unit is excluded from the result to make it easy to gauge
    completion of all peers joining the relation with existing hook tools.

    Example usage:
    log('peer {} of {} joined peer relation'
        .format(len(related_units()),
                len(list(expected_peer_units()))))

    This function will raise NotImplementedError if used with juju versions
    without goal-state support.

    :returns: iterator
    :rtype: types.GeneratorType
    :raises: NotImplementedError
    """
    if not has_juju_version("2.4.0"):
        # goal-state first appeared in 2.4.0.
        raise NotImplementedError("goal-state")
    _goal_state = goal_state()
    return (key for key in _goal_state['units']
            if '/' in key and key != local_unit())


def expected_related_units(reltype=None):
    """Get a generator for units we expect to join relation based on
    goal-state.

    Note that you can not use this function for the peer relation, take a look
    at expected_peer_units() for that.

    This function will raise KeyError if you request information for a
    relation type for which juju goal-state does not have information.  It will
    raise NotImplementedError if used with juju versions without goal-state
    support.

    Example usage:
    log('participant {} of {} joined relation {}'
        .format(len(related_units()),
                len(list(expected_related_units())),
                relation_type()))

    :param reltype: Relation type to list data for, default is to list data for
                    the realtion type we are currently executing a hook for.
    :type reltype: str
    :returns: iterator
    :rtype: types.GeneratorType
    :raises: KeyError, NotImplementedError
    """
    if not has_juju_version("2.4.4"):
        # goal-state existed in 2.4.0, but did not list individual units to
        # join a relation in 2.4.1 through 2.4.3. (LP: #1794739)
        raise NotImplementedError("goal-state relation unit count")
    reltype = reltype or relation_type()
    _goal_state = goal_state()
    return (key for key in _goal_state['relations'][reltype] if '/' in key)


@cached
def relation_for_unit(unit=None, rid=None):
    """Get the json represenation of a unit's relation"""
    unit = unit or remote_unit()
    relation = relation_get(unit=unit, rid=rid)
    for key in relation:
        if key.endswith('-list'):
            relation[key] = relation[key].split()
    relation['__unit__'] = unit
    return relation


@cached
def relations_for_id(relid=None):
    """Get relations of a specific relation ID"""
    relation_data = []
    relid = relid or relation_ids()
    for unit in related_units(relid):
        unit_data = relation_for_unit(unit, relid)
        unit_data['__relid__'] = relid
        relation_data.append(unit_data)
    return relation_data


@cached
def relations_of_type(reltype=None):
    """Get relations of a specific type"""
    relation_data = []
    reltype = reltype or relation_type()
    for relid in relation_ids(reltype):
        for relation in relations_for_id(relid):
            relation['__relid__'] = relid
            relation_data.append(relation)
    return relation_data


@cached
def metadata():
    """Get the current charm metadata.yaml contents as a python object"""
    with open(os.path.join(charm_dir(), 'metadata.yaml')) as md:
        return yaml.safe_load(md)


def _metadata_unit(unit):
    """Given the name of a unit (e.g. apache2/0), get the unit charm's
    metadata.yaml. Very similar to metadata() but allows us to inspect
    other units. Unit needs to be co-located, such as a subordinate or
    principal/primary.

    :returns: metadata.yaml as a python object.

    """
    basedir = os.sep.join(charm_dir().split(os.sep)[:-2])
    unitdir = 'unit-{}'.format(unit.replace(os.sep, '-'))
    joineddir = os.path.join(basedir, unitdir, 'charm', 'metadata.yaml')
    if not os.path.exists(joineddir):
        return None
    with open(joineddir) as md:
        return yaml.safe_load(md)


@cached
def relation_types():
    """Get a list of relation types supported by this charm"""
    rel_types = []
    md = metadata()
    for key in ('provides', 'requires', 'peers'):
        section = md.get(key)
        if section:
            rel_types.extend(section.keys())
    return rel_types


@cached
def peer_relation_id():
    '''Get the peers relation id if a peers relation has been joined, else None.'''
    md = metadata()
    section = md.get('peers')
    if section:
        for key in section:
            relids = relation_ids(key)
            if relids:
                return relids[0]
    return None


@cached
def relation_to_interface(relation_name):
    """
    Given the name of a relation, return the interface that relation uses.

    :returns: The interface name, or ``None``.
    """
    return relation_to_role_and_interface(relation_name)[1]


@cached
def relation_to_role_and_interface(relation_name):
    """
    Given the name of a relation, return the role and the name of the interface
    that relation uses (where role is one of ``provides``, ``requires``, or ``peers``).

    :returns: A tuple containing ``(role, interface)``, or ``(None, None)``.
    """
    _metadata = metadata()
    for role in ('provides', 'requires', 'peers'):
        interface = _metadata.get(role, {}).get(relation_name, {}).get('interface')
        if interface:
            return role, interface
    return None, None


@cached
def role_and_interface_to_relations(role, interface_name):
    """
    Given a role and interface name, return a list of relation names for the
    current charm that use that interface under that role (where role is one
    of ``provides``, ``requires``, or ``peers``).

    :returns: A list of relation names.
    """
    _metadata = metadata()
    results = []
    for relation_name, relation in _metadata.get(role, {}).items():
        if relation['interface'] == interface_name:
            results.append(relation_name)
    return results


@cached
def interface_to_relations(interface_name):
    """
    Given an interface, return a list of relation names for the current
    charm that use that interface.

    :returns: A list of relation names.
    """
    results = []
    for role in ('provides', 'requires', 'peers'):
        results.extend(role_and_interface_to_relations(role, interface_name))
    return results


@cached
def charm_name():
    """Get the name of the current charm as is specified on metadata.yaml"""
    return metadata().get('name')


@cached
def relations():
    """Get a nested dictionary of relation data for all related units"""
    rels = {}
    for reltype in relation_types():
        relids = {}
        for relid in relation_ids(reltype):
            units = {local_unit(): relation_get(unit=local_unit(), rid=relid)}
            for unit in related_units(relid):
                reldata = relation_get(unit=unit, rid=relid)
                units[unit] = reldata
            relids[relid] = units
        rels[reltype] = relids
    return rels


@cached
def is_relation_made(relation, keys='private-address'):
    '''
    Determine whether a relation is established by checking for
    presence of key(s).  If a list of keys is provided, they
    must all be present for the relation to be identified as made
    '''
    if isinstance(keys, str):
        keys = [keys]
    for r_id in relation_ids(relation):
        for unit in related_units(r_id):
            context = {}
            for k in keys:
                context[k] = relation_get(k, rid=r_id,
                                          unit=unit)
            if None not in context.values():
                return True
    return False


def _port_op(op_name, port, protocol="TCP"):
    """Open or close a service network port"""
    _args = [op_name]
    icmp = protocol.upper() == "ICMP"
    if icmp:
        _args.append(protocol)
    else:
        _args.append('{}/{}'.format(port, protocol))
    try:
        subprocess.check_call(_args)
    except subprocess.CalledProcessError:
        # Older Juju pre 2.3 doesn't support ICMP
        # so treat it as a no-op if it fails.
        if not icmp:
            raise


def open_port(port, protocol="TCP"):
    """Open a service network port"""
    _port_op('open-port', port, protocol)


def close_port(port, protocol="TCP"):
    """Close a service network port"""
    _port_op('close-port', port, protocol)


def open_ports(start, end, protocol="TCP"):
    """Opens a range of service network ports"""
    _args = ['open-port']
    _args.append('{}-{}/{}'.format(start, end, protocol))
    subprocess.check_call(_args)


def close_ports(start, end, protocol="TCP"):
    """Close a range of service network ports"""
    _args = ['close-port']
    _args.append('{}-{}/{}'.format(start, end, protocol))
    subprocess.check_call(_args)


def opened_ports():
    """Get the opened ports

    *Note that this will only show ports opened in a previous hook*

    :returns: Opened ports as a list of strings: ``['8080/tcp', '8081-8083/tcp']``
    """
    _args = ['opened-ports', '--format=json']
    return json.loads(subprocess.check_output(_args).decode('UTF-8'))


@cached
def unit_get(attribute):
    """Get the unit ID for the remote unit"""
    _args = ['unit-get', '--format=json', attribute]
    try:
        return json.loads(subprocess.check_output(_args).decode('UTF-8'))
    except ValueError:
        return None


def unit_public_ip():
    """Get this unit's public IP address"""
    return unit_get('public-address')


def unit_private_ip():
    """Get this unit's private IP address"""
    return unit_get('private-address')


@cached
def storage_get(attribute=None, storage_id=None):
    """Get storage attributes"""
    _args = ['storage-get', '--format=json']
    if storage_id:
        _args.extend(('-s', storage_id))
    if attribute:
        _args.append(attribute)
    try:
        return json.loads(subprocess.check_output(_args).decode('UTF-8'))
    except ValueError:
        return None


@cached
def storage_list(storage_name=None):
    """List the storage IDs for the unit"""
    _args = ['storage-list', '--format=json']
    if storage_name:
        _args.append(storage_name)
    try:
        return json.loads(subprocess.check_output(_args).decode('UTF-8'))
    except ValueError:
        return None
    except OSError as e:
        import errno
        if e.errno == errno.ENOENT:
            # storage-list does not exist
            return []
        raise


class UnregisteredHookError(Exception):
    """Raised when an undefined hook is called"""
    pass


class Hooks(object):
    """A convenient handler for hook functions.

    Example::

        hooks = Hooks()

        # register a hook, taking its name from the function name
        @hooks.hook()
        def install():
            pass  # your code here

        # register a hook, providing a custom hook name
        @hooks.hook("config-changed")
        def config_changed():
            pass  # your code here

        if __name__ == "__main__":
            # execute a hook based on the name the program is called by
            hooks.execute(sys.argv)
    """

    def __init__(self, config_save=None):
        super(Hooks, self).__init__()
        self._hooks = {}

        # For unknown reasons, we allow the Hooks constructor to override
        # config().implicit_save.
        if config_save is not None:
            config().implicit_save = config_save

    def register(self, name, function):
        """Register a hook"""
        self._hooks[name] = function

    def execute(self, args):
        """Execute a registered hook based on args[0]"""
        _run_atstart()
        hook_name = os.path.basename(args[0])
        if hook_name in self._hooks:
            try:
                self._hooks[hook_name]()
            except SystemExit as x:
                if x.code is None or x.code == 0:
                    _run_atexit()
                raise
            _run_atexit()
        else:
            raise UnregisteredHookError(hook_name)

    def hook(self, *hook_names):
        """Decorator, registering them as hooks"""
        def wrapper(decorated):
            for hook_name in hook_names:
                self.register(hook_name, decorated)
            else:
                self.register(decorated.__name__, decorated)
                if '_' in decorated.__name__:
                    self.register(
                        decorated.__name__.replace('_', '-'), decorated)
            return decorated
        return wrapper


class NoNetworkBinding(Exception):
    pass


def charm_dir():
    """Return the root directory of the current charm"""
    d = os.environ.get('JUJU_CHARM_DIR')
    if d is not None:
        return d
    return os.environ.get('CHARM_DIR')


def cmd_exists(cmd):
    """Return True if the specified cmd exists in the path"""
    return any(
        os.access(os.path.join(path, cmd), os.X_OK)
        for path in os.environ["PATH"].split(os.pathsep)
    )


@cached
@deprecate("moved to function_get()", log=log)
def action_get(key=None):
    """
    .. deprecated:: 0.20.7
       Alias for :func:`function_get`.

    Gets the value of an action parameter, or all key/value param pairs.
    """
    cmd = ['action-get']
    if key is not None:
        cmd.append(key)
    cmd.append('--format=json')
    action_data = json.loads(subprocess.check_output(cmd).decode('UTF-8'))
    return action_data


@cached
def function_get(key=None):
    """Gets the value of an action parameter, or all key/value param pairs"""
    cmd = ['function-get']
    # Fallback for older charms.
    if not cmd_exists('function-get'):
        cmd = ['action-get']

    if key is not None:
        cmd.append(key)
    cmd.append('--format=json')
    function_data = json.loads(subprocess.check_output(cmd).decode('UTF-8'))
    return function_data


@deprecate("moved to function_set()", log=log)
def action_set(values):
    """
    .. deprecated:: 0.20.7
       Alias for :func:`function_set`.

    Sets the values to be returned after the action finishes.
    """
    cmd = ['action-set']
    for k, v in list(values.items()):
        cmd.append('{}={}'.format(k, v))
    subprocess.check_call(cmd)


def function_set(values):
    """Sets the values to be returned after the function finishes"""
    cmd = ['function-set']
    # Fallback for older charms.
    if not cmd_exists('function-get'):
        cmd = ['action-set']

    for k, v in list(values.items()):
        cmd.append('{}={}'.format(k, v))
    subprocess.check_call(cmd)


@deprecate("moved to function_fail()", log=log)
def action_fail(message):
    """
    .. deprecated:: 0.20.7
       Alias for :func:`function_fail`.

    Sets the action status to failed and sets the error message.

    The results set by action_set are preserved.
    """
    subprocess.check_call(['action-fail', message])


def function_fail(message):
    """Sets the function status to failed and sets the error message.

    The results set by function_set are preserved."""
    cmd = ['function-fail']
    # Fallback for older charms.
    if not cmd_exists('function-fail'):
        cmd = ['action-fail']
    cmd.append(message)

    subprocess.check_call(cmd)


def action_name():
    """Get the name of the currently executing action."""
    return os.environ.get('JUJU_ACTION_NAME')


def function_name():
    """Get the name of the currently executing function."""
    return os.environ.get('JUJU_FUNCTION_NAME') or action_name()


def action_uuid():
    """Get the UUID of the currently executing action."""
    return os.environ.get('JUJU_ACTION_UUID')


def function_id():
    """Get the ID of the currently executing function."""
    return os.environ.get('JUJU_FUNCTION_ID') or action_uuid()


def action_tag():
    """Get the tag for the currently executing action."""
    return os.environ.get('JUJU_ACTION_TAG')


def function_tag():
    """Get the tag for the currently executing function."""
    return os.environ.get('JUJU_FUNCTION_TAG') or action_tag()


def status_set(workload_state, message, application=False):
    """Set the workload state with a message

    Use status-set to set the workload state with a message which is visible
    to the user via juju status. If the status-set command is not found then
    assume this is juju < 1.23 and juju-log the message instead.

    workload_state   -- valid juju workload state. str or WORKLOAD_STATES
    message          -- status update message
    application      -- Whether this is an application state set
    """
    bad_state_msg = '{!r} is not a valid workload state'

    if isinstance(workload_state, str):
        try:
            # Convert string to enum.
            workload_state = WORKLOAD_STATES[workload_state.upper()]
        except KeyError:
            raise ValueError(bad_state_msg.format(workload_state))

    if workload_state not in WORKLOAD_STATES:
        raise ValueError(bad_state_msg.format(workload_state))

    cmd = ['status-set']
    if application:
        cmd.append('--application')
    cmd.extend([workload_state.value, message])
    try:
        ret = subprocess.call(cmd)
        if ret == 0:
            return
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
    log_message = 'status-set failed: {} {}'.format(workload_state.value,
                                                    message)
    log(log_message, level='INFO')


def status_get():
    """Retrieve the previously set juju workload state and message

    If the status-get command is not found then assume this is juju < 1.23 and
    return 'unknown', ""

    """
    cmd = ['status-get', "--format=json", "--include-data"]
    try:
        raw_status = subprocess.check_output(cmd)
    except OSError as e:
        if e.errno == errno.ENOENT:
            return ('unknown', "")
        else:
            raise
    else:
        status = json.loads(raw_status.decode("UTF-8"))
        return (status["status"], status["message"])


def translate_exc(from_exc, to_exc):
    def inner_translate_exc1(f):
        @wraps(f)
        def inner_translate_exc2(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except from_exc:
                raise to_exc

        return inner_translate_exc2

    return inner_translate_exc1


def application_version_set(version):
    """Charm authors may trigger this command from any hook to output what
    version of the application is running. This could be a package version,
    for instance postgres version 9.5. It could also be a build number or
    version control revision identifier, for instance git sha 6fb7ba68. """

    cmd = ['application-version-set']
    cmd.append(version)
    try:
        subprocess.check_call(cmd)
    except OSError:
        log("Application Version: {}".format(version))


@translate_exc(from_exc=OSError, to_exc=NotImplementedError)
@cached
def goal_state():
    """Juju goal state values"""
    cmd = ['goal-state', '--format=json']
    return json.loads(subprocess.check_output(cmd).decode('UTF-8'))


@translate_exc(from_exc=OSError, to_exc=NotImplementedError)
def is_leader():
    """Does the current unit hold the juju leadership

    Uses juju to determine whether the current unit is the leader of its peers
    """
    cmd = ['is-leader', '--format=json']
    return json.loads(subprocess.check_output(cmd).decode('UTF-8'))


@translate_exc(from_exc=OSError, to_exc=NotImplementedError)
def leader_get(attribute=None):
    """Juju leader get value(s)"""
    cmd = ['leader-get', '--format=json'] + [attribute or '-']
    return json.loads(subprocess.check_output(cmd).decode('UTF-8'))


@translate_exc(from_exc=OSError, to_exc=NotImplementedError)
def leader_set(settings=None, **kwargs):
    """Juju leader set value(s)"""
    # Don't log secrets.
    # log("Juju leader-set '%s'" % (settings), level=DEBUG)
    cmd = ['leader-set']
    settings = settings or {}
    settings.update(kwargs)
    for k, v in settings.items():
        if v is None:
            cmd.append('{}='.format(k))
        else:
            cmd.append('{}={}'.format(k, v))
    subprocess.check_call(cmd)


@translate_exc(from_exc=OSError, to_exc=NotImplementedError)
def payload_register(ptype, klass, pid):
    """ is used while a hook is running to let Juju know that a
        payload has been started."""
    cmd = ['payload-register']
    for x in [ptype, klass, pid]:
        cmd.append(x)
    subprocess.check_call(cmd)


@translate_exc(from_exc=OSError, to_exc=NotImplementedError)
def payload_unregister(klass, pid):
    """ is used while a hook is running to let Juju know
    that a payload has been manually stopped. The <class> and <id> provided
    must match a payload that has been previously registered with juju using
    payload-register."""
    cmd = ['payload-unregister']
    for x in [klass, pid]:
        cmd.append(x)
    subprocess.check_call(cmd)


@translate_exc(from_exc=OSError, to_exc=NotImplementedError)
def payload_status_set(klass, pid, status):
    """is used to update the current status of a registered payload.
    The <class> and <id> provided must match a payload that has been previously
    registered with juju using payload-register. The <status> must be one of the
    follow: starting, started, stopping, stopped"""
    cmd = ['payload-status-set']
    for x in [klass, pid, status]:
        cmd.append(x)
    subprocess.check_call(cmd)


@translate_exc(from_exc=OSError, to_exc=NotImplementedError)
def resource_get(name):
    """used to fetch the resource path of the given name.

    <name> must match a name of defined resource in metadata.yaml

    returns either a path or False if resource not available
    """
    if not name:
        return False

    cmd = ['resource-get', name]
    try:
        return subprocess.check_output(cmd).decode('UTF-8')
    except subprocess.CalledProcessError:
        return False


@cached
def juju_version():
    """Full version string (eg. '1.23.3.1-trusty-amd64')"""
    # Per https://bugs.launchpad.net/juju-core/+bug/1455368/comments/1
    jujud = glob.glob('/var/lib/juju/tools/machine-*/jujud')[0]
    return subprocess.check_output([jujud, 'version'],
                                   universal_newlines=True).strip()


def has_juju_version(minimum_version):
    """Return True if the Juju version is at least the provided version"""
    return LooseVersion(juju_version()) >= LooseVersion(minimum_version)


_atexit = []
_atstart = []


def atstart(callback, *args, **kwargs):
    '''Schedule a callback to run before the main hook.

    Callbacks are run in the order they were added.

    This is useful for modules and classes to perform initialization
    and inject behavior. In particular:

        - Run common code before all of your hooks, such as logging
          the hook name or interesting relation data.
        - Defer object or module initialization that requires a hook
          context until we know there actually is a hook context,
          making testing easier.
        - Rather than requiring charm authors to include boilerplate to
          invoke your helper's behavior, have it run automatically if
          your object is instantiated or module imported.

    This is not at all useful after your hook framework as been launched.
    '''
    global _atstart
    _atstart.append((callback, args, kwargs))


def atexit(callback, *args, **kwargs):
    '''Schedule a callback to run on successful hook completion.

    Callbacks are run in the reverse order that they were added.'''
    _atexit.append((callback, args, kwargs))


def _run_atstart():
    '''Hook frameworks must invoke this before running the main hook body.'''
    global _atstart
    for callback, args, kwargs in _atstart:
        callback(*args, **kwargs)
    del _atstart[:]


def _run_atexit():
    '''Hook frameworks must invoke this after the main hook body has
    successfully completed. Do not invoke it if the hook fails.'''
    global _atexit
    for callback, args, kwargs in reversed(_atexit):
        callback(*args, **kwargs)
    del _atexit[:]


@translate_exc(from_exc=OSError, to_exc=NotImplementedError)
def network_get_primary_address(binding):
    '''
    Deprecated since Juju 2.3; use network_get()

    Retrieve the primary network address for a named binding

    :param binding: string. The name of a relation of extra-binding
    :return: string. The primary IP address for the named binding
    :raise: NotImplementedError if run on Juju < 2.0
    '''
    cmd = ['network-get', '--primary-address', binding]
    try:
        response = subprocess.check_output(
            cmd,
            stderr=subprocess.STDOUT).decode('UTF-8').strip()
    except CalledProcessError as e:
        if 'no network config found for binding' in e.output.decode('UTF-8'):
            raise NoNetworkBinding("No network binding for {}"
                                   .format(binding))
        else:
            raise
    return response


def network_get(endpoint, relation_id=None):
    """
    Retrieve the network details for a relation endpoint

    :param endpoint: string. The name of a relation endpoint
    :param relation_id: int. The ID of the relation for the current context.
    :return: dict. The loaded YAML output of the network-get query.
    :raise: NotImplementedError if request not supported by the Juju version.
    """
    if not has_juju_version('2.2'):
        raise NotImplementedError(juju_version())  # earlier versions require --primary-address
    if relation_id and not has_juju_version('2.3'):
        raise NotImplementedError  # 2.3 added the -r option

    cmd = ['network-get', endpoint, '--format', 'yaml']
    if relation_id:
        cmd.append('-r')
        cmd.append(relation_id)
    response = subprocess.check_output(
        cmd,
        stderr=subprocess.STDOUT).decode('UTF-8').strip()
    return yaml.safe_load(response)


def add_metric(*args, **kwargs):
    """Add metric values. Values may be expressed with keyword arguments. For
    metric names containing dashes, these may be expressed as one or more
    'key=value' positional arguments. May only be called from the collect-metrics
    hook."""
    _args = ['add-metric']
    _kvpairs = []
    _kvpairs.extend(args)
    _kvpairs.extend(['{}={}'.format(k, v) for k, v in kwargs.items()])
    _args.extend(sorted(_kvpairs))
    try:
        subprocess.check_call(_args)
        return
    except EnvironmentError as e:
        if e.errno != errno.ENOENT:
            raise
    log_message = 'add-metric failed: {}'.format(' '.join(_kvpairs))
    log(log_message, level='INFO')


def meter_status():
    """Get the meter status, if running in the meter-status-changed hook."""
    return os.environ.get('JUJU_METER_STATUS')


def meter_info():
    """Get the meter status information, if running in the meter-status-changed
    hook."""
    return os.environ.get('JUJU_METER_INFO')


def iter_units_for_relation_name(relation_name):
    """Iterate through all units in a relation

    Generator that iterates through all the units in a relation and yields
    a named tuple with rid and unit field names.

    Usage:
    data = [(u.rid, u.unit)
            for u in iter_units_for_relation_name(relation_name)]

    :param relation_name: string relation name
    :yield: Named Tuple with rid and unit field names
    """
    RelatedUnit = namedtuple('RelatedUnit', 'rid, unit')
    for rid in relation_ids(relation_name):
        for unit in related_units(rid):
            yield RelatedUnit(rid, unit)


def ingress_address(rid=None, unit=None):
    """
    Retrieve the ingress-address from a relation when available.
    Otherwise, return the private-address.

    When used on the consuming side of the relation (unit is a remote
    unit), the ingress-address is the IP address that this unit needs
    to use to reach the provided service on the remote unit.

    When used on the providing side of the relation (unit == local_unit()),
    the ingress-address is the IP address that is advertised to remote
    units on this relation. Remote units need to use this address to
    reach the local provided service on this unit.

    Note that charms may document some other method to use in
    preference to the ingress_address(), such as an address provided
    on a different relation attribute or a service discovery mechanism.
    This allows charms to redirect inbound connections to their peers
    or different applications such as load balancers.

    Usage:
    addresses = [ingress_address(rid=u.rid, unit=u.unit)
                 for u in iter_units_for_relation_name(relation_name)]

    :param rid: string relation id
    :param unit: string unit name
    :side effect: calls relation_get
    :return: string IP address
    """
    settings = relation_get(rid=rid, unit=unit)
    return (settings.get('ingress-address') or
            settings.get('private-address'))


def egress_subnets(rid=None, unit=None):
    """
    Retrieve the egress-subnets from a relation.

    This function is to be used on the providing side of the
    relation, and provides the ranges of addresses that client
    connections may come from. The result is uninteresting on
    the consuming side of a relation (unit == local_unit()).

    Returns a stable list of subnets in CIDR format.
    eg. ['192.168.1.0/24', '2001::F00F/128']

    If egress-subnets is not available, falls back to using the published
    ingress-address, or finally private-address.

    :param rid: string relation id
    :param unit: string unit name
    :side effect: calls relation_get
    :return: list of subnets in CIDR format. eg. ['192.168.1.0/24', '2001::F00F/128']
    """
    def _to_range(addr):
        if re.search(r'^(?:\d{1,3}\.){3}\d{1,3}$', addr) is not None:
            addr += '/32'
        elif ':' in addr and '/' not in addr:  # IPv6
            addr += '/128'
        return addr

    settings = relation_get(rid=rid, unit=unit)
    if 'egress-subnets' in settings:
        return [n.strip() for n in settings['egress-subnets'].split(',') if n.strip()]
    if 'ingress-address' in settings:
        return [_to_range(settings['ingress-address'])]
    if 'private-address' in settings:
        return [_to_range(settings['private-address'])]
    return []  # Should never happen


def unit_doomed(unit=None):
    """Determines if the unit is being removed from the model

    Requires Juju 2.4.1.

    :param unit: string unit name, defaults to local_unit
    :side effect: calls goal_state
    :side effect: calls local_unit
    :side effect: calls has_juju_version
    :return: True if the unit is being removed, already gone, or never existed
    """
    if not has_juju_version("2.4.1"):
        # We cannot risk blindly returning False for 'we don't know',
        # because that could cause data loss; if call sites don't
        # need an accurate answer, they likely don't need this helper
        # at all.
        # goal-state existed in 2.4.0, but did not handle removals
        # correctly until 2.4.1.
        raise NotImplementedError("is_doomed")
    if unit is None:
        unit = local_unit()
    gs = goal_state()
    units = gs.get('units', {})
    if unit not in units:
        return True
    # I don't think 'dead' units ever show up in the goal-state, but
    # check anyway in addition to 'dying'.
    return units[unit]['status'] in ('dying', 'dead')


def env_proxy_settings(selected_settings=None):
    """Get proxy settings from process environment variables.

    Get charm proxy settings from environment variables that correspond to
    juju-http-proxy, juju-https-proxy juju-no-proxy (available as of 2.4.2, see
    lp:1782236) and juju-ftp-proxy in a format suitable for passing to an
    application that reacts to proxy settings passed as environment variables.
    Some applications support lowercase or uppercase notation (e.g. curl), some
    support only lowercase (e.g. wget), there are also subjectively rare cases
    of only uppercase notation support. no_proxy CIDR and wildcard support also
    varies between runtimes and applications as there is no enforced standard.

    Some applications may connect to multiple destinations and expose config
    options that would affect only proxy settings for a specific destination
    these should be handled in charms in an application-specific manner.

    :param selected_settings: format only a subset of possible settings
    :type selected_settings: list
    :rtype: Option(None, dict[str, str])
    """
    SUPPORTED_SETTINGS = {
        'http': 'HTTP_PROXY',
        'https': 'HTTPS_PROXY',
        'no_proxy': 'NO_PROXY',
        'ftp': 'FTP_PROXY'
    }
    if selected_settings is None:
        selected_settings = SUPPORTED_SETTINGS

    selected_vars = [v for k, v in SUPPORTED_SETTINGS.items()
                     if k in selected_settings]
    proxy_settings = {}
    for var in selected_vars:
        var_val = os.getenv(var)
        if var_val:
            proxy_settings[var] = var_val
            proxy_settings[var.lower()] = var_val
        # Now handle juju-prefixed environment variables. The legacy vs new
        # environment variable usage is mutually exclusive
        charm_var_val = os.getenv('JUJU_CHARM_{}'.format(var))
        if charm_var_val:
            proxy_settings[var] = charm_var_val
            proxy_settings[var.lower()] = charm_var_val
    if 'no_proxy' in proxy_settings:
        if _contains_range(proxy_settings['no_proxy']):
            log(RANGE_WARNING, level=WARNING)
    return proxy_settings if proxy_settings else None


def _contains_range(addresses):
    """Check for cidr or wildcard domain in a string.

    Given a string comprising a comma seperated list of ip addresses
    and domain names, determine whether the string contains IP ranges
    or wildcard domains.

    :param addresses: comma seperated list of domains and ip addresses.
    :type addresses: str
    """
    return (
        # Test for cidr (e.g. 10.20.20.0/24)
        "/" in addresses or
        # Test for wildcard domains (*.foo.com or .foo.com)
        "*" in addresses or
        addresses.startswith(".") or
        ",." in addresses or
        " ." in addresses)



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import inspect

from docutils.parsers.rst import directives
from sphinx.ext.autosummary import Autosummary
from sphinx.ext.autosummary import get_import_prefixes_from_env
from sphinx.ext.autosummary import import_by_name


class AutoMemberSummary(Autosummary):
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    has_content = True
    option_spec = {
        'toctree': directives.unchanged,
        'nosignatures': directives.flag,
        'template': directives.unchanged,
    }

    def get_items(self, names):
        env = self.state.document.settings.env
        prefixes = get_import_prefixes_from_env(env)

        items = []
        prefix = ''
        shorten = ''

        def _get_items(name):
            _items = super(AutoMemberSummary, self).get_items([shorten + name])
            for dn, sig, summary, rn in _items:
                items.append(('%s%s' % (prefix, dn), sig, summary, rn))

        for name in names:
            if '~' in name:
                prefix, name = name.split('~')
                shorten = '~'
            else:
                prefix = ''
                shorten = ''

            try:
                real_name, obj, parent, _ = import_by_name(name, prefixes=prefixes)
            except ImportError:
                self.warn('failed to import %s' % name)
                continue

            if not inspect.ismodule(obj):
                _get_items(name)
                continue

            for member in dir(obj):
                if member.startswith('_'):
                    continue
                mobj = getattr(obj, member)
                if hasattr(mobj, '__module__'):
                    if not mobj.__module__.startswith(real_name):
                        continue  # skip imported classes & functions
                elif hasattr(mobj, '__name__'):
                    if not mobj.__name__.startswith(real_name):
                        continue  # skip imported modules
                else:
                    continue  # skip instances
                _get_items('%s.%s' % (name, member))

        return items


def setup(app):
    app.add_directive('automembersummary', AutoMemberSummary)



from mock import patch
from testtools import TestCase

from charmhelpers.contrib import ssl


class HelpersTest(TestCase):
    @patch('subprocess.check_call')
    def test_generate_selfsigned_dict(self, mock_call):
        subject = {"country": "UK",
                   "locality": "my_locality",
                   "state": "my_state",
                   "organization": "my_organization",
                   "organizational_unit": "my_unit",
                   "cn": "mysite.example.com",
                   "email": "me@example.com"
                   }

        ssl.generate_selfsigned("mykey.key", "mycert.crt", subject=subject)
        mock_call.assert_called_with(['/usr/bin/openssl', 'req', '-new',
                                      '-newkey', 'rsa:1024', '-days', '365',
                                      '-nodes', '-x509', '-keyout',
                                      'mykey.key', '-out', 'mycert.crt',
                                      '-subj',
                                      '/C=UK/ST=my_state/L=my_locality'
                                      '/O=my_organization/OU=my_unit'
                                      '/CN=mysite.example.com'
                                      '/emailAddress=me@example.com']
                                     )

    @patch('charmhelpers.core.hookenv.log')
    def test_generate_selfsigned_failure(self, mock_log):
        # This is NOT enough, functino requires cn key
        subject = {"country": "UK",
                   "locality": "my_locality"}

        result = ssl.generate_selfsigned("mykey.key", "mycert.crt", subject=subject)
        self.assertFalse(result)

    @patch('subprocess.check_call')
    def test_generate_selfsigned_file(self, mock_call):
        ssl.generate_selfsigned("mykey.key", "mycert.crt", config="test.cnf")
        mock_call.assert_called_with(['/usr/bin/openssl', 'req', '-new',
                                      '-newkey', 'rsa:1024', '-days', '365',
                                      '-nodes', '-x509', '-keyout',
                                      'mykey.key', '-out', 'mycert.crt',
                                      '-config', 'test.cnf'])

    @patch('subprocess.check_call')
    def test_generate_selfsigned_cn_key(self, mock_call):
        ssl.generate_selfsigned("mykey.key", "mycert.crt", keysize="2048", cn="mysite.example.com")
        mock_call.assert_called_with(['/usr/bin/openssl', 'req', '-new',
                                      '-newkey', 'rsa:2048', '-days', '365',
                                      '-nodes', '-x509', '-keyout',
                                      'mykey.key', '-out', 'mycert.crt',
                                      '-subj', '/CN=mysite.example.com'])



import os

from unittest import TestCase
from mock import (
    MagicMock,
    patch,
    mock_open,
    Mock,
    ANY
)
from charmhelpers.fetch import (
    archiveurl,
    UnhandledSource,
)

import six
if six.PY3:
    from urllib.parse import urlparse
    from urllib.error import URLError
else:
    from urllib2 import URLError
    from urlparse import urlparse


class ArchiveUrlFetchHandlerTest(TestCase):

    def setUp(self):
        super(ArchiveUrlFetchHandlerTest, self).setUp()
        self.valid_urls = (
            "http://example.com/foo.tar.gz",
            "http://example.com/foo.tgz",
            "http://example.com/foo.tar.bz2",
            "http://example.com/foo.tbz2",
            "http://example.com/foo.zip",
            "http://example.com/foo.zip?bar=baz&x=y#whee",
            "ftp://example.com/foo.tar.gz",
            "https://example.com/foo.tgz",
            "file://example.com/foo.tar.bz2",
        )
        self.invalid_urls = (
            "git://example.com/foo.tar.gz",
            "http://example.com/foo",
            "http://example.com/foobar=baz&x=y#tar.gz",
            "http://example.com/foobar?h=baz.zip",
            "bzr+ssh://example.com/foo.tar.gz",
            "lp:example/foo.tgz",
            "file//example.com/foo.tar.bz2",
            "garbage",
        )
        self.fh = archiveurl.ArchiveUrlFetchHandler()

    def test_handles_archive_urls(self):
        for url in self.valid_urls:
            result = self.fh.can_handle(url)
            self.assertEqual(result, True, url)
        for url in self.invalid_urls:
            result = self.fh.can_handle(url)
            self.assertNotEqual(result, True, url)

    @patch('charmhelpers.fetch.archiveurl.urlopen')
    def test_downloads(self, _urlopen):
        for url in self.valid_urls:
            response = MagicMock()
            response.read.return_value = "bar"
            _urlopen.return_value = response

            _open = mock_open()
            with patch('charmhelpers.fetch.archiveurl.open',
                       _open, create=True):
                self.fh.download(url, "foo")

            response.read.assert_called_with()
            _open.assert_called_once_with("foo", 'wb')
            _open().write.assert_called_with("bar")

    @patch('charmhelpers.fetch.archiveurl.check_hash')
    @patch('charmhelpers.fetch.archiveurl.mkdir')
    @patch('charmhelpers.fetch.archiveurl.extract')
    def test_installs(self, _extract, _mkdir, _check_hash):
        self.fh.download = MagicMock()

        for url in self.valid_urls:
            filename = urlparse(url).path
            dest = os.path.join('foo', 'fetched', os.path.basename(filename))
            _extract.return_value = dest
            with patch.dict('os.environ', {'CHARM_DIR': 'foo'}):
                where = self.fh.install(url, checksum='deadbeef')
            self.fh.download.assert_called_with(url, dest)
            _extract.assert_called_with(dest, None)
            _check_hash.assert_called_with(dest, 'deadbeef', 'sha1')
            self.assertEqual(where, dest)
            _check_hash.reset_mock()

        url = "http://www.example.com/archive.tar.gz"

        self.fh.download.side_effect = URLError('fail')
        with patch.dict('os.environ', {'CHARM_DIR': 'foo'}):
            self.assertRaises(UnhandledSource, self.fh.install, url)

        self.fh.download.side_effect = OSError('fail')
        with patch.dict('os.environ', {'CHARM_DIR': 'foo'}):
            self.assertRaises(UnhandledSource, self.fh.install, url)

    @patch('charmhelpers.fetch.archiveurl.check_hash')
    @patch('charmhelpers.fetch.archiveurl.mkdir')
    @patch('charmhelpers.fetch.archiveurl.extract')
    def test_install_with_hash_in_url(self, _extract, _mkdir, _check_hash):
        self.fh.download = MagicMock()
        url = "file://example.com/foo.tar.bz2#sha512=beefdead"
        with patch.dict('os.environ', {'CHARM_DIR': 'foo'}):
            self.fh.install(url)
        _check_hash.assert_called_with(ANY, 'beefdead', 'sha512')

    @patch('charmhelpers.fetch.archiveurl.mkdir')
    @patch('charmhelpers.fetch.archiveurl.extract')
    def test_install_with_duplicate_hash_in_url(self, _extract, _mkdir):
        self.fh.download = MagicMock()
        url = "file://example.com/foo.tar.bz2#sha512=a&sha512=b"
        with patch.dict('os.environ', {'CHARM_DIR': 'foo'}):
            with self.assertRaisesRegexp(
                    TypeError, "Expected 1 hash value, not 2"):
                self.fh.install(url)

    @patch('charmhelpers.fetch.archiveurl.urlretrieve')
    @patch('charmhelpers.fetch.archiveurl.check_hash')
    def test_download_and_validate(self, vfmock, urlmock):
        urlmock.return_value = ('/tmp/tmpebM9Hv', Mock())
        dlurl = 'http://example.com/foo.tgz'
        dlhash = '988881adc9fc3655077dc2d4d757d480b5ea0e11'
        self.fh.download_and_validate(dlurl, dlhash)
        vfmock.assert_called_with('/tmp/tmpebM9Hv', dlhash, 'sha1')



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compatibility with the nrpe-external-master charm"""
# Copyright 2012 Canonical Ltd.
#
# Authors:
#  Matthew Wedgwood <matthew.wedgwood@canonical.com>

import subprocess
import pwd
import grp
import os
import glob
import shutil
import re
import shlex
import yaml

from charmhelpers.core.hookenv import (
    config,
    hook_name,
    local_unit,
    log,
    relation_get,
    relation_ids,
    relation_set,
    relations_of_type,
)

from charmhelpers.core.host import service
from charmhelpers.core import host

# This module adds compatibility with the nrpe-external-master and plain nrpe
# subordinate charms. To use it in your charm:
#
# 1. Update metadata.yaml
#
#   provides:
#     (...)
#     nrpe-external-master:
#       interface: nrpe-external-master
#       scope: container
#
#   and/or
#
#   provides:
#     (...)
#     local-monitors:
#       interface: local-monitors
#       scope: container

#
# 2. Add the following to config.yaml
#
#    nagios_context:
#      default: "juju"
#      type: string
#      description: |
#        Used by the nrpe subordinate charms.
#        A string that will be prepended to instance name to set the host name
#        in nagios. So for instance the hostname would be something like:
#            juju-myservice-0
#        If you're running multiple environments with the same services in them
#        this allows you to differentiate between them.
#    nagios_servicegroups:
#      default: ""
#      type: string
#      description: |
#        A comma-separated list of nagios servicegroups.
#        If left empty, the nagios_context will be used as the servicegroup
#
# 3. Add custom checks (Nagios plugins) to files/nrpe-external-master
#
# 4. Update your hooks.py with something like this:
#
#    from charmsupport.nrpe import NRPE
#    (...)
#    def update_nrpe_config():
#        nrpe_compat = NRPE()
#        nrpe_compat.add_check(
#            shortname = "myservice",
#            description = "Check MyService",
#            check_cmd = "check_http -w 2 -c 10 http://localhost"
#            )
#        nrpe_compat.add_check(
#            "myservice_other",
#            "Check for widget failures",
#            check_cmd = "/srv/myapp/scripts/widget_check"
#            )
#        nrpe_compat.write()
#
#    def config_changed():
#        (...)
#        update_nrpe_config()
#
#    def nrpe_external_master_relation_changed():
#        update_nrpe_config()
#
#    def local_monitors_relation_changed():
#        update_nrpe_config()
#
# 4.a If your charm is a subordinate charm set primary=False
#
#    from charmsupport.nrpe import NRPE
#    (...)
#    def update_nrpe_config():
#        nrpe_compat = NRPE(primary=False)
#
# 5. ln -s hooks.py nrpe-external-master-relation-changed
#    ln -s hooks.py local-monitors-relation-changed


class CheckException(Exception):
    pass


class Check(object):
    shortname_re = '[A-Za-z0-9-_.@]+$'
    service_template = ("""
#---------------------------------------------------
# This file is Juju managed
#---------------------------------------------------
define service {{
    use                             active-service
    host_name                       {nagios_hostname}
    service_description             {nagios_hostname}[{shortname}] """
                        """{description}
    check_command                   check_nrpe!{command}
    servicegroups                   {nagios_servicegroup}
}}
""")

    def __init__(self, shortname, description, check_cmd):
        super(Check, self).__init__()
        # XXX: could be better to calculate this from the service name
        if not re.match(self.shortname_re, shortname):
            raise CheckException("shortname must match {}".format(
                Check.shortname_re))
        self.shortname = shortname
        self.command = "check_{}".format(shortname)
        # Note: a set of invalid characters is defined by the
        # Nagios server config
        # The default is: illegal_object_name_chars=`~!$%^&*"|'<>?,()=
        self.description = description
        self.check_cmd = self._locate_cmd(check_cmd)

    def _get_check_filename(self):
        return os.path.join(NRPE.nrpe_confdir, '{}.cfg'.format(self.command))

    def _get_service_filename(self, hostname):
        return os.path.join(NRPE.nagios_exportdir,
                            'service__{}_{}.cfg'.format(hostname, self.command))

    def _locate_cmd(self, check_cmd):
        search_path = (
            '/usr/lib/nagios/plugins',
            '/usr/local/lib/nagios/plugins',
        )
        parts = shlex.split(check_cmd)
        for path in search_path:
            if os.path.exists(os.path.join(path, parts[0])):
                command = os.path.join(path, parts[0])
                if len(parts) > 1:
                    command += " " + " ".join(parts[1:])
                return command
        log('Check command not found: {}'.format(parts[0]))
        return ''

    def _remove_service_files(self):
        if not os.path.exists(NRPE.nagios_exportdir):
            return
        for f in os.listdir(NRPE.nagios_exportdir):
            if f.endswith('_{}.cfg'.format(self.command)):
                os.remove(os.path.join(NRPE.nagios_exportdir, f))

    def remove(self, hostname):
        nrpe_check_file = self._get_check_filename()
        if os.path.exists(nrpe_check_file):
            os.remove(nrpe_check_file)
        self._remove_service_files()

    def write(self, nagios_context, hostname, nagios_servicegroups):
        nrpe_check_file = self._get_check_filename()
        with open(nrpe_check_file, 'w') as nrpe_check_config:
            nrpe_check_config.write("# check {}\n".format(self.shortname))
            if nagios_servicegroups:
                nrpe_check_config.write(
                    "# The following header was added automatically by juju\n")
                nrpe_check_config.write(
                    "# Modifying it will affect nagios monitoring and alerting\n")
                nrpe_check_config.write(
                    "# servicegroups: {}\n".format(nagios_servicegroups))
            nrpe_check_config.write("command[{}]={}\n".format(
                self.command, self.check_cmd))

        if not os.path.exists(NRPE.nagios_exportdir):
            log('Not writing service config as {} is not accessible'.format(
                NRPE.nagios_exportdir))
        else:
            self.write_service_config(nagios_context, hostname,
                                      nagios_servicegroups)

    def write_service_config(self, nagios_context, hostname,
                             nagios_servicegroups):
        self._remove_service_files()

        templ_vars = {
            'nagios_hostname': hostname,
            'nagios_servicegroup': nagios_servicegroups,
            'description': self.description,
            'shortname': self.shortname,
            'command': self.command,
        }
        nrpe_service_text = Check.service_template.format(**templ_vars)
        nrpe_service_file = self._get_service_filename(hostname)
        with open(nrpe_service_file, 'w') as nrpe_service_config:
            nrpe_service_config.write(str(nrpe_service_text))

    def run(self):
        subprocess.call(self.check_cmd)


class NRPE(object):
    nagios_logdir = '/var/log/nagios'
    nagios_exportdir = '/var/lib/nagios/export'
    nrpe_confdir = '/etc/nagios/nrpe.d'
    homedir = '/var/lib/nagios'  # home dir provided by nagios-nrpe-server

    def __init__(self, hostname=None, primary=True):
        super(NRPE, self).__init__()
        self.config = config()
        self.primary = primary
        self.nagios_context = self.config['nagios_context']
        if 'nagios_servicegroups' in self.config and self.config['nagios_servicegroups']:
            self.nagios_servicegroups = self.config['nagios_servicegroups']
        else:
            self.nagios_servicegroups = self.nagios_context
        self.unit_name = local_unit().replace('/', '-')
        if hostname:
            self.hostname = hostname
        else:
            nagios_hostname = get_nagios_hostname()
            if nagios_hostname:
                self.hostname = nagios_hostname
            else:
                self.hostname = "{}-{}".format(self.nagios_context, self.unit_name)
        self.checks = []
        # Iff in an nrpe-external-master relation hook, set primary status
        relation = relation_ids('nrpe-external-master')
        if relation:
            log("Setting charm primary status {}".format(primary))
            for rid in relation:
                relation_set(relation_id=rid, relation_settings={'primary': self.primary})
        self.remove_check_queue = set()

    def add_check(self, *args, **kwargs):
        shortname = None
        if kwargs.get('shortname') is None:
            if len(args) > 0:
                shortname = args[0]
        else:
            shortname = kwargs['shortname']

        self.checks.append(Check(*args, **kwargs))
        try:
            self.remove_check_queue.remove(shortname)
        except KeyError:
            pass

    def remove_check(self, *args, **kwargs):
        if kwargs.get('shortname') is None:
            raise ValueError('shortname of check must be specified')

        # Use sensible defaults if they're not specified - these are not
        # actually used during removal, but they're required for constructing
        # the Check object; check_disk is chosen because it's part of the
        # nagios-plugins-basic package.
        if kwargs.get('check_cmd') is None:
            kwargs['check_cmd'] = 'check_disk'
        if kwargs.get('description') is None:
            kwargs['description'] = ''

        check = Check(*args, **kwargs)
        check.remove(self.hostname)
        self.remove_check_queue.add(kwargs['shortname'])

    def write(self):
        try:
            nagios_uid = pwd.getpwnam('nagios').pw_uid
            nagios_gid = grp.getgrnam('nagios').gr_gid
        except Exception:
            log("Nagios user not set up, nrpe checks not updated")
            return

        if not os.path.exists(NRPE.nagios_logdir):
            os.mkdir(NRPE.nagios_logdir)
            os.chown(NRPE.nagios_logdir, nagios_uid, nagios_gid)

        nrpe_monitors = {}
        monitors = {"monitors": {"remote": {"nrpe": nrpe_monitors}}}
        for nrpecheck in self.checks:
            nrpecheck.write(self.nagios_context, self.hostname,
                            self.nagios_servicegroups)
            nrpe_monitors[nrpecheck.shortname] = {
                "command": nrpecheck.command,
            }

        # update-status hooks are configured to firing every 5 minutes by
        # default. When nagios-nrpe-server is restarted, the nagios server
        # reports checks failing causing unnecessary alerts. Let's not restart
        # on update-status hooks.
        if not hook_name() == 'update-status':
            service('restart', 'nagios-nrpe-server')

        monitor_ids = relation_ids("local-monitors") + \
            relation_ids("nrpe-external-master")
        for rid in monitor_ids:
            reldata = relation_get(unit=local_unit(), rid=rid)
            if 'monitors' in reldata:
                # update the existing set of monitors with the new data
                old_monitors = yaml.safe_load(reldata['monitors'])
                old_nrpe_monitors = old_monitors['monitors']['remote']['nrpe']
                # remove keys that are in the remove_check_queue
                old_nrpe_monitors = {k: v for k, v in old_nrpe_monitors.items()
                                     if k not in self.remove_check_queue}
                # update/add nrpe_monitors
                old_nrpe_monitors.update(nrpe_monitors)
                old_monitors['monitors']['remote']['nrpe'] = old_nrpe_monitors
                # write back to the relation
                relation_set(relation_id=rid, monitors=yaml.dump(old_monitors))
            else:
                # write a brand new set of monitors, as no existing ones.
                relation_set(relation_id=rid, monitors=yaml.dump(monitors))

        self.remove_check_queue.clear()


def get_nagios_hostcontext(relation_name='nrpe-external-master'):
    """
    Query relation with nrpe subordinate, return the nagios_host_context

    :param str relation_name: Name of relation nrpe sub joined to
    """
    for rel in relations_of_type(relation_name):
        if 'nagios_host_context' in rel:
            return rel['nagios_host_context']


def get_nagios_hostname(relation_name='nrpe-external-master'):
    """
    Query relation with nrpe subordinate, return the nagios_hostname

    :param str relation_name: Name of relation nrpe sub joined to
    """
    for rel in relations_of_type(relation_name):
        if 'nagios_hostname' in rel:
            return rel['nagios_hostname']


def get_nagios_unit_name(relation_name='nrpe-external-master'):
    """
    Return the nagios unit name prepended with host_context if needed

    :param str relation_name: Name of relation nrpe sub joined to
    """
    host_context = get_nagios_hostcontext(relation_name)
    if host_context:
        unit = "%s:%s" % (host_context, local_unit())
    else:
        unit = local_unit()
    return unit


def add_init_service_checks(nrpe, services, unit_name, immediate_check=True):
    """
    Add checks for each service in list

    :param NRPE nrpe: NRPE object to add check to
    :param list services: List of services to check
    :param str unit_name: Unit name to use in check description
    :param bool immediate_check: For sysv init, run the service check immediately
    """
    for svc in services:
        # Don't add a check for these services from neutron-gateway
        if svc in ['ext-port', 'os-charm-phy-nic-mtu']:
            next

        upstart_init = '/etc/init/%s.conf' % svc
        sysv_init = '/etc/init.d/%s' % svc

        if host.init_is_systemd():
            nrpe.add_check(
                shortname=svc,
                description='process check {%s}' % unit_name,
                check_cmd='check_systemd.py %s' % svc
            )
        elif os.path.exists(upstart_init):
            nrpe.add_check(
                shortname=svc,
                description='process check {%s}' % unit_name,
                check_cmd='check_upstart_job %s' % svc
            )
        elif os.path.exists(sysv_init):
            cronpath = '/etc/cron.d/nagios-service-check-%s' % svc
            checkpath = '%s/service-check-%s.txt' % (nrpe.homedir, svc)
            croncmd = (
                '/usr/local/lib/nagios/plugins/check_exit_status.pl '
                '-e -s /etc/init.d/%s status' % svc
            )
            cron_file = '*/5 * * * * root %s > %s\n' % (croncmd, checkpath)
            f = open(cronpath, 'w')
            f.write(cron_file)
            f.close()
            nrpe.add_check(
                shortname=svc,
                description='service check {%s}' % unit_name,
                check_cmd='check_status_file.py -f %s' % checkpath,
            )
            # if /var/lib/nagios doesn't exist open(checkpath, 'w') will fail
            # (LP: #1670223).
            if immediate_check and os.path.isdir(nrpe.homedir):
                f = open(checkpath, 'w')
                subprocess.call(
                    croncmd.split(),
                    stdout=f,
                    stderr=subprocess.STDOUT
                )
                f.close()
                os.chmod(checkpath, 0o644)


def copy_nrpe_checks(nrpe_files_dir=None):
    """
    Copy the nrpe checks into place

    """
    NAGIOS_PLUGINS = '/usr/local/lib/nagios/plugins'
    if nrpe_files_dir is None:
        # determine if "charmhelpers" is in CHARMDIR or CHARMDIR/hooks
        for segment in ['.', 'hooks']:
            nrpe_files_dir = os.path.abspath(os.path.join(
                os.getenv('CHARM_DIR'),
                segment,
                'charmhelpers',
                'contrib',
                'openstack',
                'files'))
            if os.path.isdir(nrpe_files_dir):
                break
        else:
            raise RuntimeError("Couldn't find charmhelpers directory")
    if not os.path.exists(NAGIOS_PLUGINS):
        os.makedirs(NAGIOS_PLUGINS)
    for fname in glob.glob(os.path.join(nrpe_files_dir, "check_*")):
        if os.path.isfile(fname):
            shutil.copy2(fname,
                         os.path.join(NAGIOS_PLUGINS, os.path.basename(fname)))


def add_haproxy_checks(nrpe, unit_name):
    """
    Add checks for each service in list

    :param NRPE nrpe: NRPE object to add check to
    :param str unit_name: Unit name to use in check description
    """
    nrpe.add_check(
        shortname='haproxy_servers',
        description='Check HAProxy {%s}' % unit_name,
        check_cmd='check_haproxy.sh')
    nrpe.add_check(
        shortname='haproxy_queue',
        description='Check HAProxy queue depth {%s}' % unit_name,
        check_cmd='check_haproxy_queue_depth.sh')


def remove_deprecated_check(nrpe, deprecated_services):
    """
    Remove checks fro deprecated services in list

    :param nrpe: NRPE object to remove check from
    :type nrpe: NRPE
    :param deprecated_services: List of deprecated services that are removed
    :type deprecated_services: list
    """
    for dep_svc in deprecated_services:
        log('Deprecated service: {}'.format(dep_svc))
        nrpe.remove_check(shortname=dep_svc)



#!/usr/bin/python
#
# Copyright 2017 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
from charmhelpers.fetch import apt_install
from charmhelpers.contrib.openstack.context import IdentityServiceContext
from charmhelpers.core.hookenv import (
    log,
    ERROR,
)


def get_api_suffix(api_version):
    """Return the formatted api suffix for the given version
    @param api_version: version of the keystone endpoint
    @returns the api suffix formatted according to the given api
    version
    """
    return 'v2.0' if api_version in (2, "2", "2.0") else 'v3'


def format_endpoint(schema, addr, port, api_version):
    """Return a formatted keystone endpoint
    @param schema: http or https
    @param addr: ipv4/ipv6 host of the keystone service
    @param port: port of the keystone service
    @param api_version: 2 or 3
    @returns a fully formatted keystone endpoint
    """
    return '{}://{}:{}/{}/'.format(schema, addr, port,
                                   get_api_suffix(api_version))


def get_keystone_manager(endpoint, api_version, **kwargs):
    """Return a keystonemanager for the correct API version

    @param endpoint: the keystone endpoint to point client at
    @param api_version: version of the keystone api the client should use
    @param kwargs: token or username/tenant/password information
    @returns keystonemanager class used for interrogating keystone
    """
    if api_version == 2:
        return KeystoneManager2(endpoint, **kwargs)
    if api_version == 3:
        return KeystoneManager3(endpoint, **kwargs)
    raise ValueError('No manager found for api version {}'.format(api_version))


def get_keystone_manager_from_identity_service_context():
    """Return a keystonmanager generated from a
    instance of charmhelpers.contrib.openstack.context.IdentityServiceContext
    @returns keystonamenager instance
   """
    context = IdentityServiceContext()()
    if not context:
        msg = "Identity service context cannot be generated"
        log(msg, level=ERROR)
        raise ValueError(msg)

    endpoint = format_endpoint(context['service_protocol'],
                               context['service_host'],
                               context['service_port'],
                               context['api_version'])

    if context['api_version'] in (2, "2.0"):
        api_version = 2
    else:
        api_version = 3

    return get_keystone_manager(endpoint, api_version,
                                username=context['admin_user'],
                                password=context['admin_password'],
                                tenant_name=context['admin_tenant_name'])


class KeystoneManager(object):

    def resolve_service_id(self, service_name=None, service_type=None):
        """Find the service_id of a given service"""
        services = [s._info for s in self.api.services.list()]

        service_name = service_name.lower()
        for s in services:
            name = s['name'].lower()
            if service_type and service_name:
                if (service_name == name and service_type == s['type']):
                    return s['id']
            elif service_name and service_name == name:
                return s['id']
            elif service_type and service_type == s['type']:
                return s['id']
        return None

    def service_exists(self, service_name=None, service_type=None):
        """Determine if the given service exists on the service list"""
        return self.resolve_service_id(service_name, service_type) is not None


class KeystoneManager2(KeystoneManager):

    def __init__(self, endpoint, **kwargs):
        try:
            from keystoneclient.v2_0 import client
            from keystoneclient.auth.identity import v2
            from keystoneclient import session
        except ImportError:
            if six.PY2:
                apt_install(["python-keystoneclient"], fatal=True)
            else:
                apt_install(["python3-keystoneclient"], fatal=True)

            from keystoneclient.v2_0 import client
            from keystoneclient.auth.identity import v2
            from keystoneclient import session

        self.api_version = 2

        token = kwargs.get("token", None)
        if token:
            api = client.Client(endpoint=endpoint, token=token)
        else:
            auth = v2.Password(username=kwargs.get("username"),
                               password=kwargs.get("password"),
                               tenant_name=kwargs.get("tenant_name"),
                               auth_url=endpoint)
            sess = session.Session(auth=auth)
            api = client.Client(session=sess)

        self.api = api


class KeystoneManager3(KeystoneManager):

    def __init__(self, endpoint, **kwargs):
        try:
            from keystoneclient.v3 import client
            from keystoneclient.auth import token_endpoint
            from keystoneclient import session
            from keystoneclient.auth.identity import v3
        except ImportError:
            if six.PY2:
                apt_install(["python-keystoneclient"], fatal=True)
            else:
                apt_install(["python3-keystoneclient"], fatal=True)

            from keystoneclient.v3 import client
            from keystoneclient.auth import token_endpoint
            from keystoneclient import session
            from keystoneclient.auth.identity import v3

        self.api_version = 3

        token = kwargs.get("token", None)
        if token:
            auth = token_endpoint.Token(endpoint=endpoint,
                                        token=token)
            sess = session.Session(auth=auth)
        else:
            auth = v3.Password(auth_url=endpoint,
                               user_id=kwargs.get("username"),
                               password=kwargs.get("password"),
                               project_id=kwargs.get("tenant_name"))
            sess = session.Session(auth=auth)

        self.api = client.Client(session=sess)



# Copyright 2019 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import uuid

from . import utils


class SimpleOVSDB(object):
    """Simple interface to OVSDB through the use of command line tools.

    OVS and OVN is managed through a set of databases.  These databases have
    similar command line tools to manage them.  We make use of the similarity
    to provide a generic class that can be used to manage them.

    The OpenvSwitch project does provide a Python API, but on the surface it
    appears to be a bit too involved for our simple use case.

    Examples:
    sbdb = SimpleOVSDB('ovn-sbctl')
    for chs in sbdb.chassis:
        print(chs)

    ovsdb = SimpleOVSDB('ovs-vsctl')
    for br in ovsdb.bridge:
        if br['name'] == 'br-test':
            ovsdb.bridge.set(br['uuid'], 'external_ids:charm', 'managed')
    """

    # For validation we keep a complete map of currently known good tool and
    # table combinations.  This requires maintenance down the line whenever
    # upstream adds things that downstream wants, and the cost of maintaining
    # that will most likely be lower then the cost of finding the needle in
    # the haystack whenever downstream code misspells something.
    _tool_table_map = {
        'ovs-vsctl': (
            'autoattach',
            'bridge',
            'ct_timeout_policy',
            'ct_zone',
            'controller',
            'datapath',
            'flow_sample_collector_set',
            'flow_table',
            'ipfix',
            'interface',
            'manager',
            'mirror',
            'netflow',
            'open_vswitch',
            'port',
            'qos',
            'queue',
            'ssl',
            'sflow',
        ),
        'ovn-nbctl': (
            'acl',
            'address_set',
            'connection',
            'dhcp_options',
            'dns',
            'forwarding_group',
            'gateway_chassis',
            'ha_chassis',
            'ha_chassis_group',
            'load_balancer',
            'load_balancer_health_check',
            'logical_router',
            'logical_router_policy',
            'logical_router_port',
            'logical_router_static_route',
            'logical_switch',
            'logical_switch_port',
            'meter',
            'meter_band',
            'nat',
            'nb_global',
            'port_group',
            'qos',
            'ssl',
        ),
        'ovn-sbctl': (
            'address_set',
            'chassis',
            'connection',
            'controller_event',
            'dhcp_options',
            'dhcpv6_options',
            'dns',
            'datapath_binding',
            'encap',
            'gateway_chassis',
            'ha_chassis',
            'ha_chassis_group',
            'igmp_group',
            'ip_multicast',
            'logical_flow',
            'mac_binding',
            'meter',
            'meter_band',
            'multicast_group',
            'port_binding',
            'port_group',
            'rbac_permission',
            'rbac_role',
            'sb_global',
            'ssl',
            'service_monitor',
        ),
    }

    def __init__(self, tool):
        """SimpleOVSDB constructor.

        :param tool: Which tool with database commands to operate on.
                     Usually one of `ovs-vsctl`, `ovn-nbctl`, `ovn-sbctl`
        :type tool: str
        """
        if tool not in self._tool_table_map:
            raise RuntimeError(
                'tool must be one of "{}"'.format(self._tool_table_map.keys()))
        self._tool = tool

    def __getattr__(self, table):
        if table not in self._tool_table_map[self._tool]:
            raise AttributeError(
                'table "{}" not known for use with "{}"'
                .format(table, self._tool))
        return self.Table(self._tool, table)

    class Table(object):
        """Methods to interact with contents of OVSDB tables.

        NOTE: At the time of this writing ``find`` is the only command
        line argument to OVSDB manipulating tools that actually supports
        JSON output.
        """

        def __init__(self, tool, table):
            """SimpleOVSDBTable constructor.

            :param table: Which table to operate on
            :type table: str
            """
            self._tool = tool
            self._table = table

        def _find_tbl(self, condition=None):
            """Run and parse output of OVSDB `find` command.

            :param condition: An optional RFC 7047 5.1 match condition
            :type condition: Optional[str]
            :returns: Dictionary with data
            :rtype: Dict[str, any]
            """
            # When using json formatted output to OVS commands Internal OVSDB
            # notation may occur that require further deserializing.
            # Reference: https://tools.ietf.org/html/rfc7047#section-5.1
            ovs_type_cb_map = {
                'uuid': uuid.UUID,
                # FIXME sets also appear to sometimes contain type/value tuples
                'set': list,
                'map': dict,
            }
            cmd = [self._tool, '-f', 'json', 'find', self._table]
            if condition:
                cmd.append(condition)
            output = utils._run(*cmd)
            data = json.loads(output)
            for row in data['data']:
                values = []
                for col in row:
                    if isinstance(col, list):
                        f = ovs_type_cb_map.get(col[0], str)
                        values.append(f(col[1]))
                    else:
                        values.append(col)
                yield dict(zip(data['headings'], values))

        def __iter__(self):
            return self._find_tbl()

        def clear(self, rec, col):
            utils._run(self._tool, 'clear', self._table, rec, col)

        def find(self, condition):
            return self._find_tbl(condition=condition)

        def remove(self, rec, col, value):
            utils._run(self._tool, 'remove', self._table, rec, col, value)

        def set(self, rec, col, value):
            utils._run(self._tool, 'set', self._table, rec,
                       '{}={}'.format(col, value))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = 'Jorge Niedbalski <niedbalski@ubuntu.com>'

import os
import subprocess


def sed(filename, before, after, flags='g'):
    """
    Search and replaces the given pattern on filename.

    :param filename: relative or absolute file path.
    :param before: expression to be replaced (see 'man sed')
    :param after: expression to replace with (see 'man sed')
    :param flags: sed-compatible regex flags in example, to make
    the  search and replace case insensitive, specify ``flags="i"``.
    The ``g`` flag is always specified regardless, so you do not
    need to remember to include it when overriding this parameter.
    :returns: If the sed command exit code was zero then return,
    otherwise raise CalledProcessError.
    """
    expression = r's/{0}/{1}/{2}'.format(before,
                                         after, flags)

    return subprocess.check_call(["sed", "-i", "-r", "-e",
                                  expression,
                                  os.path.expanduser(filename)])



from testtools import TestCase, skipIf
from mock import patch, MagicMock, call
import six

import charmhelpers.contrib.openstack.audits as audits
import charmhelpers.contrib.openstack.audits.openstack_security_guide as guide


@skipIf(six.PY2, 'Audits only support Python3')
class AuditTestCase(TestCase):

    @patch('charmhelpers.contrib.openstack.audits._audits', {})
    def test_wrapper(self):
        variables = {
            'guard_called': False,
            'test_run': False,
        }

        def should_run(audit_options):
            variables['guard_called'] = True
            return True

        @audits.audit(should_run)
        def test(options):
            variables['test_run'] = True

        audits.run({})
        self.assertTrue(variables['guard_called'])
        self.assertTrue(variables['test_run'])
        self.assertEqual(audits._audits['test'], audits.Audit(test, (should_run,)))

    @patch('charmhelpers.contrib.openstack.audits._audits', {})
    def test_wrapper_not_run(self):
        variables = {
            'guard_called': False,
            'test_run': False,
        }

        def should_run(audit_options):
            variables['guard_called'] = True
            return False

        @audits.audit(should_run)
        def test(options):
            variables['test_run'] = True

        audits.run({})
        self.assertTrue(variables['guard_called'])
        self.assertFalse(variables['test_run'])
        self.assertEqual(audits._audits['test'], audits.Audit(test, (should_run,)))

    @patch('charmhelpers.contrib.openstack.audits._audits', {})
    def test_duplicate_audit(self):
        def should_run(audit_options):
            return True

        @audits.audit(should_run)
        def test(options):
            pass

        try:
            # Again!
            #
            # Both of the following '#noqa's are to prevent flake8 from
            # noticing the duplicate function `test`  The intent in this test
            # is for the audits.audit to pick up on the duplicate function.
            @audits.audit(should_run)  # noqa
            def test(options):         # noqa
                pass
        except RuntimeError as e:
            self.assertEqual("Test name 'test' used more than once", e.args[0])
            return
        self.assertTrue(False, "Duplicate audit should raise an exception")

    @patch('charmhelpers.contrib.openstack.audits._audits', {})
    def test_non_callable_filter(self):
        try:
            # Again!
            @audits.audit(3)
            def test(options):
                pass
        except RuntimeError as e:
            self.assertEqual("Configuration includes non-callable filters: [3]", e.args[0])
            return
        self.assertTrue(False, "Duplicate audit should raise an exception")

    @patch('charmhelpers.contrib.openstack.audits._audits', {})
    def test_exclude_config(self):
        variables = {
            'test_run': False,
        }

        @audits.audit()
        def test(options):
            variables['test_run'] = True

        audits.run({'excludes': ['test']})
        self.assertFalse(variables['test_run'])


class AuditsTestCase(TestCase):

    @patch('charmhelpers.contrib.openstack.audits.cmp_pkgrevno')
    def test_since_package_less(self, _cmp_pkgrevno):
        _cmp_pkgrevno.return_value = 1

        verifier = audits.since_package('test', '12.0.0')
        self.assertEqual(verifier(), True)

    @patch('charmhelpers.contrib.openstack.audits.cmp_pkgrevno')
    def test_since_package_greater(self, _cmp_pkgrevno):
        _cmp_pkgrevno.return_value = -1

        verifier = audits.since_package('test', '14.0.0')
        self.assertEqual(verifier(), False)

    @patch('charmhelpers.contrib.openstack.audits.cmp_pkgrevno')
    def test_since_package_equal(self, _cmp_pkgrevno):
        _cmp_pkgrevno.return_value = 0

        verifier = audits.since_package('test', '13.0.0')
        self.assertEqual(verifier(), True)

    @patch('charmhelpers.contrib.openstack.utils.get_os_codename_package')
    def test_since_openstack_less(self, _get_os_codename_package):
        _get_os_codename_package.return_value = "icehouse"

        verifier = audits.since_openstack_release('test', 'mitaka')
        self.assertEqual(verifier(), False)

    @patch('charmhelpers.contrib.openstack.utils.get_os_codename_package')
    def test_since_openstack_greater(self, _get_os_codename_package):
        _get_os_codename_package.return_value = "rocky"

        verifier = audits.since_openstack_release('test', 'queens')
        self.assertEqual(verifier(), True)

    @patch('charmhelpers.contrib.openstack.utils.get_os_codename_package')
    def test_since_openstack_equal(self, _get_os_codename_package):
        _get_os_codename_package.return_value = "mitaka"

        verifier = audits.since_openstack_release('test', 'mitaka')
        self.assertEqual(verifier(), True)

    @patch('charmhelpers.contrib.openstack.utils.get_os_codename_package')
    def test_before_openstack_less(self, _get_os_codename_package):
        _get_os_codename_package.return_value = "icehouse"

        verifier = audits.before_openstack_release('test', 'mitaka')
        self.assertEqual(verifier(), True)

    @patch('charmhelpers.contrib.openstack.utils.get_os_codename_package')
    def test_before_openstack_greater(self, _get_os_codename_package):
        _get_os_codename_package.return_value = "rocky"

        verifier = audits.before_openstack_release('test', 'queens')
        self.assertEqual(verifier(), False)

    @patch('charmhelpers.contrib.openstack.utils.get_os_codename_package')
    def test_before_openstack_equal(self, _get_os_codename_package):
        _get_os_codename_package.return_value = "mitaka"

        verifier = audits.before_openstack_release('test', 'mitaka')
        self.assertEqual(verifier(), False)

    @patch('charmhelpers.contrib.openstack.audits.cmp_pkgrevno')
    def test_before_package_less(self, _cmp_pkgrevno):
        _cmp_pkgrevno.return_value = 1

        verifier = audits.before_package('test', '12.0.0')
        self.assertEqual(verifier(), False)

    @patch('charmhelpers.contrib.openstack.audits.cmp_pkgrevno')
    def test_before_package_greater(self, _cmp_pkgrevno):
        _cmp_pkgrevno.return_value = -1

        verifier = audits.before_package('test', '14.0.0')
        self.assertEqual(verifier(), True)

    @patch('charmhelpers.contrib.openstack.audits.cmp_pkgrevno')
    def test_before_package_equal(self, _cmp_pkgrevno):
        _cmp_pkgrevno.return_value = 0

        verifier = audits.before_package('test', '13.0.0')
        self.assertEqual(verifier(), False)

    def test_is_audit_type_empty(self):
        verifier = audits.is_audit_type(audits.AuditType.OpenStackSecurityGuide)
        self.assertEqual(verifier({}), False)

    def test_is_audit_type(self):
        verifier = audits.is_audit_type(audits.AuditType.OpenStackSecurityGuide)
        self.assertEqual(verifier({'audit_type': audits.AuditType.OpenStackSecurityGuide}), True)


@skipIf(six.PY2, 'Audits only support Python3')
class OpenstackSecurityGuideTestCase(TestCase):

    @patch('configparser.ConfigParser')
    def test_internal_config_parser_is_not_strict(self, _config_parser):
        parser = MagicMock()
        _config_parser.return_value = parser
        guide._config_ini('test')
        _config_parser.assert_called_with(strict=False)
        parser.read.assert_called_with('test')

    @patch('charmhelpers.contrib.openstack.audits.openstack_security_guide._stat')
    def test_internal_validate_file_ownership(self, _stat):
        _stat.return_value = guide.Ownership('test_user', 'test_group', '600')
        guide._validate_file_ownership('test_user', 'test_group', 'test-file-name')
        _stat.assert_called_with('test-file-name')
        pass

    @patch('charmhelpers.contrib.openstack.audits.openstack_security_guide._stat')
    def test_internal_validate_file_mode(self, _stat):
        _stat.return_value = guide.Ownership('test_user', 'test_group', '600')
        guide._validate_file_mode('600', 'test-file-name')
        _stat.assert_called_with('test-file-name')
        pass

    @patch('os.path.isfile')
    @patch('charmhelpers.contrib.openstack.audits.openstack_security_guide._validate_file_mode')
    def test_validate_file_permissions_defaults(self, _validate_mode, _is_file):
        _is_file.return_value = True
        config = {
            'files': {
                'test': {}
            }
        }
        guide.validate_file_permissions(config)
        _validate_mode.assert_called_once_with('600', 'test', False)

    @patch('os.path.isfile')
    @patch('charmhelpers.contrib.openstack.audits.openstack_security_guide._validate_file_mode')
    def test_validate_file_permissions(self, _validate_mode, _is_file):
        _is_file.return_value = True
        config = {
            'files': {
                'test': {
                    'mode': '777'
                }
            }
        }
        guide.validate_file_permissions(config)
        _validate_mode.assert_called_once_with('777', 'test', False)

    @patch('glob.glob')
    @patch('os.path.isfile')
    @patch('charmhelpers.contrib.openstack.audits.openstack_security_guide._validate_file_mode')
    def test_validate_file_permissions_glob(self, _validate_mode, _is_file, _glob):
        _glob.return_value = ['test']
        _is_file.return_value = True
        config = {
            'files': {
                '*': {
                    'mode': '777'
                }
            }
        }
        guide.validate_file_permissions(config)
        _validate_mode.assert_called_once_with('777', 'test', False)

    @patch('os.path.isfile')
    @patch('charmhelpers.contrib.openstack.audits.openstack_security_guide._validate_file_ownership')
    def test_validate_file_ownership_defaults(self, _validate_owner, _is_file):
        _is_file.return_value = True
        config = {
            'files': {
                'test': {}
            }
        }
        guide.validate_file_ownership(config)
        _validate_owner.assert_called_once_with('root', 'root', 'test', False)

    @patch('os.path.isfile')
    @patch('charmhelpers.contrib.openstack.audits.openstack_security_guide._validate_file_ownership')
    def test_validate_file_ownership(self, _validate_owner, _is_file):
        _is_file.return_value = True
        config = {
            'files': {
                'test': {
                    'owner': 'test-user',
                    'group': 'test-group',
                }
            }
        }
        guide.validate_file_ownership(config)
        _validate_owner.assert_called_once_with('test-user', 'test-group', 'test', False)

    @patch('glob.glob')
    @patch('os.path.isfile')
    @patch('charmhelpers.contrib.openstack.audits.openstack_security_guide._validate_file_ownership')
    def test_validate_file_ownership_glob(self, _validate_owner, _is_file, _glob):
        _glob.return_value = ['test']
        _is_file.return_value = True
        config = {
            'files': {
                '*': {
                    'owner': 'test-user',
                    'group': 'test-group',
                }
            }
        }
        guide.validate_file_ownership(config)
        _validate_owner.assert_called_once_with('test-user', 'test-group', 'test', False)

    @patch('charmhelpers.contrib.openstack.audits.openstack_security_guide._config_section')
    def test_validate_uses_keystone(self, _config_section):
        _config_section.side_effect = [None, {
            'auth_strategy': 'keystone',
        }]
        guide.validate_uses_keystone({})
        _config_section.assert_has_calls([call({}, 'api'), call({}, 'DEFAULT')])

    @patch('charmhelpers.contrib.openstack.audits.openstack_security_guide._config_section')
    def test_validate_uses_tls_for_keystone(self, _config_section):
        _config_section.return_value = {
            'auth_uri': 'https://10.10.10.10',
        }
        guide.validate_uses_tls_for_keystone({})
        _config_section.assert_called_with({}, 'keystone_authtoken')

    @patch('charmhelpers.contrib.openstack.audits.openstack_security_guide._config_section')
    def test_validate_uses_tls_for_glance(self, _config_section):
        _config_section.return_value = {
            'api_servers': 'https://10.10.10.10',
        }
        guide.validate_uses_tls_for_glance({})
        _config_section.assert_called_with({}, 'glance')



# Copyright 2019 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import subprocess
import uuid

from . import utils


OVN_RUNDIR = '/var/run/ovn'
OVN_SYSCONFDIR = '/etc/ovn'


def ovn_appctl(target, args, rundir=None, use_ovs_appctl=False):
    """Run ovn/ovs-appctl for target with args and return output.

    :param target: Name of daemon to contact.  Unless target begins with '/',
                   `ovn-appctl` looks for a pidfile and will build the path to
                   a /var/run/ovn/target.pid.ctl for you.
    :type target: str
    :param args: Command and arguments to pass to `ovn-appctl`
    :type args: Tuple[str, ...]
    :param rundir: Override path to sockets
    :type rundir: Optional[str]
    :param use_ovs_appctl: The ``ovn-appctl`` command appeared in OVN 20.03,
                           set this to True to use ``ovs-appctl`` instead.
    :type use_ovs_appctl: bool
    :returns: Output from command
    :rtype: str
    :raises: subprocess.CalledProcessError
    """
    # NOTE(fnordahl): The ovsdb-server processes for the OVN databases use a
    # non-standard naming scheme for their daemon control socket and we need
    # to pass the full path to the socket.
    if target in ('ovnnb_db', 'ovnsb_db',):
        target = os.path.join(rundir or OVN_RUNDIR, target + '.ctl')

    if use_ovs_appctl:
        tool = 'ovs-appctl'
    else:
        tool = 'ovn-appctl'

    return utils._run(tool, '-t', target, *args)


class OVNClusterStatus(object):

    def __init__(self, name, cluster_id, server_id, address, status, role,
                 term, leader, vote, election_timer, log,
                 entries_not_yet_committed, entries_not_yet_applied,
                 connections, servers):
        """Initialize and populate OVNClusterStatus object.

        Use class initializer so we can define types in a compatible manner.

        :param name: Name of schema used for database
        :type name: str
        :param cluster_id: UUID of cluster
        :type cluster_id: uuid.UUID
        :param server_id: UUID of server
        :type server_id: uuid.UUID
        :param address: OVSDB connection method
        :type address: str
        :param status: Status text
        :type status: str
        :param role: Role of server
        :type role: str
        :param term: Election term
        :type term: int
        :param leader: Short form UUID of leader
        :type leader: str
        :param vote: Vote
        :type vote: str
        :param election_timer: Current value of election timer
        :type election_timer: int
        :param log: Log
        :type log: str
        :param entries_not_yet_committed: Entries not yet committed
        :type entries_not_yet_committed: int
        :param entries_not_yet_applied: Entries not yet applied
        :type entries_not_yet_applied: int
        :param connections: Connections
        :type connections: str
        :param servers: Servers in the cluster
            [('0ea6', 'ssl:192.0.2.42:6643')]
        :type servers: List[Tuple[str,str]]
        """
        self.name = name
        self.cluster_id = cluster_id
        self.server_id = server_id
        self.address = address
        self.status = status
        self.role = role
        self.term = term
        self.leader = leader
        self.vote = vote
        self.election_timer = election_timer
        self.log = log
        self.entries_not_yet_committed = entries_not_yet_committed
        self.entries_not_yet_applied = entries_not_yet_applied
        self.connections = connections
        self.servers = servers

    def __eq__(self, other):
        return (
            self.name == other.name and
            self.cluster_id == other.cluster_id and
            self.server_id == other.server_id and
            self.address == other.address and
            self.status == other.status and
            self.role == other.role and
            self.term == other.term and
            self.leader == other.leader and
            self.vote == other.vote and
            self.election_timer == other.election_timer and
            self.log == other.log and
            self.entries_not_yet_committed == other.entries_not_yet_committed and
            self.entries_not_yet_applied == other.entries_not_yet_applied and
            self.connections == other.connections and
            self.servers == other.servers)

    @property
    def is_cluster_leader(self):
        """Retrieve status information from clustered OVSDB.

        :returns: Whether target is cluster leader
        :rtype: bool
        """
        return self.leader == 'self'


def cluster_status(target, schema=None, use_ovs_appctl=False, rundir=None):
    """Retrieve status information from clustered OVSDB.

    :param target: Usually one of 'ovsdb-server', 'ovnnb_db', 'ovnsb_db', can
                   also be full path to control socket.
    :type target: str
    :param schema: Database schema name, deduced from target if not provided
    :type schema: Optional[str]
    :param use_ovs_appctl: The ``ovn-appctl`` command appeared in OVN 20.03,
                           set this to True to use ``ovs-appctl`` instead.
    :type use_ovs_appctl: bool
    :param rundir: Override path to sockets
    :type rundir: Optional[str]
    :returns: cluster status data object
    :rtype: OVNClusterStatus
    :raises: subprocess.CalledProcessError, KeyError, RuntimeError
    """
    schema_map = {
        'ovnnb_db': 'OVN_Northbound',
        'ovnsb_db': 'OVN_Southbound',
    }
    if schema and schema not in schema_map.keys():
        raise RuntimeError('Unknown schema provided: "{}"'.format(schema))

    status = {}
    k = ''
    for line in ovn_appctl(target,
                           ('cluster/status', schema or schema_map[target]),
                           rundir=rundir,
                           use_ovs_appctl=use_ovs_appctl).splitlines():
        if k and line.startswith(' '):
            # there is no key which means this is a instance of a multi-line/
            # multi-value item, populate the List which is already stored under
            # the key.
            if k == 'servers':
                status[k].append(
                    tuple(line.replace(')', '').lstrip().split()[0:4:3]))
            else:
                status[k].append(line.lstrip())
        elif ':' in line:
            # this is a line with a key
            k, v = line.split(':', 1)
            k = k.lower()
            k = k.replace(' ', '_')
            if v:
                # this is a line with both key and value
                if k in ('cluster_id', 'server_id',):
                    v = v.replace('(', '')
                    v = v.replace(')', '')
                    status[k] = tuple(v.split())
                else:
                    status[k] = v.lstrip()
            else:
                # this is a line with only key which means a multi-line/
                # multi-value item.  Store key as List which will be
                # populated on subsequent iterations.
                status[k] = []
    return OVNClusterStatus(
        status['name'],
        uuid.UUID(status['cluster_id'][1]),
        uuid.UUID(status['server_id'][1]),
        status['address'],
        status['status'],
        status['role'],
        int(status['term']),
        status['leader'],
        status['vote'],
        int(status['election_timer']),
        status['log'],
        int(status['entries_not_yet_committed']),
        int(status['entries_not_yet_applied']),
        status['connections'],
        status['servers'])


def is_northd_active():
    """Query `ovn-northd` for active status.

    Note that the active status information for ovn-northd is available for
    OVN 20.03 and onward.

    :returns: True if local `ovn-northd` instance is active, False otherwise
    :rtype: bool
    """
    try:
        for line in ovn_appctl('ovn-northd', ('status',)).splitlines():
            if line.startswith('Status:') and 'active' in line:
                return True
    except subprocess.CalledProcessError:
        pass
    return False



import mock

import charmhelpers.contrib.network.ovs as ovs

import tests.utils as test_utils


# NOTE(fnordahl): some functions drectly under the ``contrib.network.ovs``
# module have their unit tests in the ``test_ovs.py`` module in the
# ``tests.contrib.network`` package.


class TestOVS(test_utils.BaseTestCase):

    def test__dict_to_vsctl_set(self):
        indata = {
            'key': 'value',
            'otherkey': {
                'nestedkey': 'nestedvalue',
            },
        }
        # due to varying Dict ordering depending on Python version we need
        # to be a bit elaborate rather than comparing result directly
        result1 = ('--', 'set', 'aTable', 'anItem', 'key=value')
        result2 = ('--', 'set', 'aTable', 'anItem',
                   'otherkey:nestedkey=nestedvalue')
        for setcmd in ovs._dict_to_vsctl_set(indata, 'aTable', 'anItem'):
            self.assertTrue(setcmd == result1 or setcmd == result2)

    def test_add_bridge(self):
        self.patch_object(ovs.subprocess, 'check_call')
        self.patch_object(ovs, 'log')
        ovs.add_bridge('test')
        self.check_call.assert_called_once_with([
            "ovs-vsctl", "--", "--may-exist",
            "add-br", 'test'])
        self.assertTrue(self.log.call_count == 1)

        self.check_call.reset_mock()
        self.log.reset_mock()
        ovs.add_bridge('test', datapath_type='netdev')
        self.check_call.assert_called_with([
            "ovs-vsctl", "--", "--may-exist",
            "add-br", 'test', "--", "set",
            "bridge", "test", "datapath_type=netdev",
        ])
        self.assertTrue(self.log.call_count == 2)

        self.check_call.reset_mock()
        ovs.add_bridge('test', exclusive=True)
        self.check_call.assert_called_once_with([
            "ovs-vsctl", "--", "add-br", 'test'])

        self.check_call.reset_mock()
        self.patch_object(ovs, '_dict_to_vsctl_set')
        self._dict_to_vsctl_set.return_value = [['--', 'fakeextradata']]
        ovs.add_bridge('test', brdata={'fakeinput': None})
        self._dict_to_vsctl_set.assert_called_once_with(
            {'fakeinput': None}, 'bridge', 'test')
        self.check_call.assert_called_once_with([
            'ovs-vsctl', '--', '--may-exist', 'add-br', 'test',
            '--', 'fakeextradata'])

    def test_add_bridge_port(self):
        self.patch_object(ovs.subprocess, 'check_call')
        self.patch_object(ovs, 'log')
        ovs.add_bridge_port('test', 'eth1')
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', '--may-exist', 'add-port',
                       'test', 'eth1']),
            mock.call(['ip', 'link', 'set', 'eth1', 'up']),
            mock.call(['ip', 'link', 'set', 'eth1', 'promisc', 'off'])
        ])
        self.assertTrue(self.log.call_count == 1)

        self.check_call.reset_mock()
        self.log.reset_mock()
        ovs.add_bridge_port('test', 'eth1', promisc=True)
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', '--may-exist', 'add-port',
                       'test', 'eth1']),
            mock.call(['ip', 'link', 'set', 'eth1', 'up']),
            mock.call(['ip', 'link', 'set', 'eth1', 'promisc', 'on'])
        ])
        self.assertTrue(self.log.call_count == 1)

        self.check_call.reset_mock()
        self.log.reset_mock()
        ovs.add_bridge_port('test', 'eth1', promisc=None)
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', '--may-exist', 'add-port',
                       'test', 'eth1']),
            mock.call(['ip', 'link', 'set', 'eth1', 'up']),
        ])
        self.assertTrue(self.log.call_count == 1)

        self.check_call.reset_mock()
        ovs.add_bridge_port('test', 'eth1', exclusive=True, linkup=False)
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', 'add-port', 'test', 'eth1']),
            mock.call(['ip', 'link', 'set', 'eth1', 'promisc', 'off'])
        ])

        self.check_call.reset_mock()
        self.patch_object(ovs, '_dict_to_vsctl_set')
        self._dict_to_vsctl_set.return_value = [['--', 'fakeextradata']]
        ovs.add_bridge_port('test', 'eth1', ifdata={'fakeinput': None})
        self._dict_to_vsctl_set.assert_called_once_with(
            {'fakeinput': None}, 'Interface', 'eth1')
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', '--may-exist', 'add-port',
                       'test', 'eth1', '--', 'fakeextradata']),
            mock.call(['ip', 'link', 'set', 'eth1', 'up']),
            mock.call(['ip', 'link', 'set', 'eth1', 'promisc', 'off'])
        ])
        self._dict_to_vsctl_set.reset_mock()
        self.check_call.reset_mock()
        ovs.add_bridge_port('test', 'eth1', portdata={'fakeportinput': None})
        self._dict_to_vsctl_set.assert_called_once_with(
            {'fakeportinput': None}, 'Port', 'eth1')
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', '--may-exist', 'add-port',
                       'test', 'eth1', '--', 'fakeextradata']),
            mock.call(['ip', 'link', 'set', 'eth1', 'up']),
            mock.call(['ip', 'link', 'set', 'eth1', 'promisc', 'off'])
        ])

    def test_ovs_appctl(self):
        self.patch_object(ovs.subprocess, 'check_output')
        ovs.ovs_appctl('ovs-vswitchd', ('ofproto/list',))
        self.check_output.assert_called_once_with(
            ['ovs-appctl', '-t', 'ovs-vswitchd', 'ofproto/list'],
            universal_newlines=True)

    def test_add_bridge_bond(self):
        self.patch_object(ovs.subprocess, 'check_call')
        self.patch_object(ovs, '_dict_to_vsctl_set')
        self._dict_to_vsctl_set.return_value = [['--', 'fakekey=fakevalue']]
        portdata = {
            'bond-mode': 'balance-tcp',
            'lacp': 'active',
            'other-config': {
                'lacp-time': 'fast',
            },
        }
        ifdatamap = {
            'eth0': {
                'type': 'dpdk',
                'mtu-request': '9000',
                'options': {
                    'dpdk-devargs': '0000:01:00.0',
                },
            },
            'eth1': {
                'type': 'dpdk',
                'mtu-request': '9000',
                'options': {
                    'dpdk-devargs': '0000:02:00.0',
                },
            },
        }
        ovs.add_bridge_bond('br-ex', 'bond42', ['eth0', 'eth1'],
                            portdata, ifdatamap)
        self._dict_to_vsctl_set.assert_has_calls([
            mock.call(portdata, 'port', 'bond42'),
            mock.call(ifdatamap['eth0'], 'Interface', 'eth0'),
            mock.call(ifdatamap['eth1'], 'Interface', 'eth1'),
        ], any_order=True)
        self.check_call.assert_called_once_with([
            'ovs-vsctl',
            '--', '--may-exist', 'add-bond', 'br-ex', 'bond42', 'eth0', 'eth1',
            '--', 'fakekey=fakevalue',
            '--', 'fakekey=fakevalue',
            '--', 'fakekey=fakevalue'])



# Copyright 2019 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide a subset of the ``python-apt`` module API.

Data collection is done through subprocess calls to ``apt-cache`` and
``dpkg-query`` commands.

The main purpose for this module is to avoid dependency on the
``python-apt`` python module.

The indicated python module is a wrapper around the ``apt`` C++ library
which is tightly connected to the version of the distribution it was
shipped on.  It is not developed in a backward/forward compatible manner.

This in turn makes it incredibly hard to distribute as a wheel for a piece
of python software that supports a span of distro releases [0][1].

Upstream feedback like [2] does not give confidence in this ever changing,
so with this we get rid of the dependency.

0: https://github.com/juju-solutions/layer-basic/pull/135
1: https://bugs.launchpad.net/charm-octavia/+bug/1824112
2: https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=845330#10
"""

import locale
import os
import subprocess
import sys


class _container(dict):
    """Simple container for attributes."""
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


class Package(_container):
    """Simple container for package attributes."""


class Version(_container):
    """Simple container for version attributes."""


class Cache(object):
    """Simulation of ``apt_pkg`` Cache object."""
    def __init__(self, progress=None):
        pass

    def __contains__(self, package):
        try:
            pkg = self.__getitem__(package)
            return pkg is not None
        except KeyError:
            return False

    def __getitem__(self, package):
        """Get information about a package from apt and dpkg databases.

        :param package: Name of package
        :type package: str
        :returns: Package object
        :rtype: object
        :raises: KeyError, subprocess.CalledProcessError
        """
        apt_result = self._apt_cache_show([package])[package]
        apt_result['name'] = apt_result.pop('package')
        pkg = Package(apt_result)
        dpkg_result = self._dpkg_list([package]).get(package, {})
        current_ver = None
        installed_version = dpkg_result.get('version')
        if installed_version:
            current_ver = Version({'ver_str': installed_version})
        pkg.current_ver = current_ver
        pkg.architecture = dpkg_result.get('architecture')
        return pkg

    def _dpkg_list(self, packages):
        """Get data from system dpkg database for package.

        :param packages: Packages to get data from
        :type packages: List[str]
        :returns: Structured data about installed packages, keys like
                  ``dpkg-query --list``
        :rtype: dict
        :raises: subprocess.CalledProcessError
        """
        pkgs = {}
        cmd = ['dpkg-query', '--list']
        cmd.extend(packages)
        if locale.getlocale() == (None, None):
            # subprocess calls out to locale.getpreferredencoding(False) to
            # determine encoding.  Workaround for Trusty where the
            # environment appears to not be set up correctly.
            locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')
        try:
            output = subprocess.check_output(cmd,
                                             stderr=subprocess.STDOUT,
                                             universal_newlines=True)
        except subprocess.CalledProcessError as cp:
            # ``dpkg-query`` may return error and at the same time have
            # produced useful output, for example when asked for multiple
            # packages where some are not installed
            if cp.returncode != 1:
                raise
            output = cp.output
        headings = []
        for line in output.splitlines():
            if line.startswith('||/'):
                headings = line.split()
                headings.pop(0)
                continue
            elif (line.startswith('|') or line.startswith('+') or
                  line.startswith('dpkg-query:')):
                continue
            else:
                data = line.split(None, 4)
                status = data.pop(0)
                if status != 'ii':
                    continue
                pkg = {}
                pkg.update({k.lower(): v for k, v in zip(headings, data)})
                if 'name' in pkg:
                    pkgs.update({pkg['name']: pkg})
        return pkgs

    def _apt_cache_show(self, packages):
        """Get data from system apt cache for package.

        :param packages: Packages to get data from
        :type packages: List[str]
        :returns: Structured data about package, keys like
                  ``apt-cache show``
        :rtype: dict
        :raises: subprocess.CalledProcessError
        """
        pkgs = {}
        cmd = ['apt-cache', 'show', '--no-all-versions']
        cmd.extend(packages)
        if locale.getlocale() == (None, None):
            # subprocess calls out to locale.getpreferredencoding(False) to
            # determine encoding.  Workaround for Trusty where the
            # environment appears to not be set up correctly.
            locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')
        try:
            output = subprocess.check_output(cmd,
                                             stderr=subprocess.STDOUT,
                                             universal_newlines=True)
            previous = None
            pkg = {}
            for line in output.splitlines():
                if not line:
                    if 'package' in pkg:
                        pkgs.update({pkg['package']: pkg})
                        pkg = {}
                    continue
                if line.startswith(' '):
                    if previous and previous in pkg:
                        pkg[previous] += os.linesep + line.lstrip()
                    continue
                if ':' in line:
                    kv = line.split(':', 1)
                    key = kv[0].lower()
                    if key == 'n':
                        continue
                    previous = key
                    pkg.update({key: kv[1].lstrip()})
        except subprocess.CalledProcessError as cp:
            # ``apt-cache`` returns 100 if none of the packages asked for
            # exist in the apt cache.
            if cp.returncode != 100:
                raise
        return pkgs


class Config(_container):
    def __init__(self):
        super(Config, self).__init__(self._populate())

    def _populate(self):
        cfgs = {}
        cmd = ['apt-config', 'dump']
        output = subprocess.check_output(cmd,
                                         stderr=subprocess.STDOUT,
                                         universal_newlines=True)
        for line in output.splitlines():
            if not line.startswith("CommandLine"):
                k, v = line.split(" ", 1)
                cfgs[k] = v.strip(";").strip("\"")

        return cfgs


# Backwards compatibility with old apt_pkg module
sys.modules[__name__].config = Config()


def init():
    """Compability shim that does nothing."""
    pass


def upstream_version(version):
    """Extracts upstream version from a version string.

    Upstream reference: https://salsa.debian.org/apt-team/apt/blob/master/
                                apt-pkg/deb/debversion.cc#L259

    :param version: Version string
    :type version: str
    :returns: Upstream version
    :rtype: str
    """
    if version:
        version = version.split(':')[-1]
        version = version.split('-')[0]
    return version


def version_compare(a, b):
    """Compare the given versions.

    Call out to ``dpkg`` to make sure the code doing the comparison is
    compatible with what the ``apt`` library would do.  Mimic the return
    values.

    Upstream reference:
    https://apt-team.pages.debian.net/python-apt/library/apt_pkg.html
            ?highlight=version_compare#apt_pkg.version_compare

    :param a: version string
    :type a: str
    :param b: version string
    :type b: str
    :returns: >0 if ``a`` is greater than ``b``, 0 if a equals b,
              <0 if ``a`` is smaller than ``b``
    :rtype: int
    :raises: subprocess.CalledProcessError, RuntimeError
    """
    for op in ('gt', 1), ('eq', 0), ('lt', -1):
        try:
            subprocess.check_call(['dpkg', '--compare-versions',
                                   a, op[0], b],
                                  stderr=subprocess.STDOUT,
                                  universal_newlines=True)
            return op[1]
        except subprocess.CalledProcessError as cp:
            if cp.returncode == 1:
                continue
            raise
    else:
        raise RuntimeError('Unable to compare "{}" and "{}", according to '
                           'our logic they are neither greater, equal nor '
                           'less than each other.')




import os
import unittest

from mock import patch, call, MagicMock

import charmhelpers.contrib.openstack.templating as templating

from jinja2.exceptions import TemplateNotFound

import six
if not six.PY3:
    builtin_open = '__builtin__.open'
else:
    builtin_open = 'builtins.open'


class FakeContextGenerator(object):
    interfaces = None

    def set(self, interfaces, context):
        self.interfaces = interfaces
        self.context = context

    def __call__(self):
        return self.context


class FakeLoader(object):
    def set(self, template):
        self.template = template

    def get(self, name):
        return self.template


class MockFSLoader(object):
    def __init__(self, dirs):
        self.searchpath = [dirs]


class MockChoiceLoader(object):
    def __init__(self, loaders):
        self.loaders = loaders


def MockTemplate():
    templ = MagicMock()
    templ.render = MagicMock()
    return templ


class TemplatingTests(unittest.TestCase):
    def setUp(self):
        path = os.path.dirname(__file__)
        self.loader = FakeLoader()
        self.context = FakeContextGenerator()

        self.addCleanup(patch.object(templating, 'apt_install').start().stop())
        self.addCleanup(patch.object(templating, 'log').start().stop())

        templating.FileSystemLoader = MockFSLoader
        templating.ChoiceLoader = MockChoiceLoader
        templating.Environment = MagicMock

        self.renderer = templating.OSConfigRenderer(templates_dir=path,
                                                    openstack_release='folsom')

    @patch.object(templating, 'apt_install')
    def test_initializing_a_render_ensures_jinja2_present(self, apt):
        '''Creatinga new renderer object installs jinja2 if needed'''
        # temp. undo the patching from setUp
        templating.FileSystemLoader = None
        templating.ChoiceLoader = None
        templating.Environment = None
        templating.OSConfigRenderer(templates_dir='/tmp',
                                    openstack_release='foo')
        templating.FileSystemLoader = MockFSLoader
        templating.ChoiceLoader = MockChoiceLoader
        templating.Environment = MagicMock
        if six.PY2:
            apt.assert_called_with('python-jinja2')
        else:
            apt.assert_called_with('python3-jinja2')

    def test_create_renderer_invalid_templates_dir(self):
        '''Ensure OSConfigRenderer checks templates_dir'''
        self.assertRaises(templating.OSConfigException,
                          templating.OSConfigRenderer,
                          templates_dir='/tmp/foooo0',
                          openstack_release='grizzly')

    def test_render_unregistered_config(self):
        '''Ensure cannot render an unregistered config file'''
        self.assertRaises(templating.OSConfigException,
                          self.renderer.render,
                          config_file='/tmp/foo')

    def test_write_unregistered_config(self):
        '''Ensure cannot write an unregistered config file'''
        self.assertRaises(templating.OSConfigException,
                          self.renderer.write,
                          config_file='/tmp/foo')

    def test_render_complete_context(self):
        '''It renders a template when provided a complete context'''
        self.loader.set('{{ foo }}')
        self.context.set(interfaces=['fooservice'], context={'foo': 'bar'})
        self.renderer.register('/tmp/foo', [self.context])
        with patch.object(self.renderer, '_get_template') as _get_t:
            fake_tmpl = MockTemplate()
            _get_t.return_value = fake_tmpl
            self.renderer.render('/tmp/foo')
            fake_tmpl.render.assert_called_with(self.context())
        self.assertIn('fooservice', self.renderer.complete_contexts())

    def test_render_incomplete_context_with_template(self):
        '''It renders a template when provided an incomplete context'''
        self.context.set(interfaces=['fooservice'], context={})
        self.renderer.register('/tmp/foo', [self.context])
        with patch.object(self.renderer, '_get_template') as _get_t:
            fake_tmpl = MockTemplate()
            _get_t.return_value = fake_tmpl
            self.renderer.render('/tmp/foo')
            fake_tmpl.render.assert_called_with({})
            self.assertNotIn('fooservice', self.renderer.complete_contexts())

    def test_render_template_registered_but_not_found(self):
        '''It loads a template by basename of config file first'''
        path = os.path.dirname(__file__)
        renderer = templating.OSConfigRenderer(templates_dir=path,
                                               openstack_release='folsom')
        e = TemplateNotFound('')
        renderer._get_template = MagicMock()
        renderer._get_template.side_effect = e
        renderer.register('/etc/nova/nova.conf', contexts=[])
        self.assertRaises(
            TemplateNotFound, renderer.render, '/etc/nova/nova.conf')

    def test_render_template_by_basename_first(self):
        '''It loads a template by basename of config file first'''
        path = os.path.dirname(__file__)
        renderer = templating.OSConfigRenderer(templates_dir=path,
                                               openstack_release='folsom')
        renderer._get_template = MagicMock()
        renderer.register('/etc/nova/nova.conf', contexts=[])
        renderer.render('/etc/nova/nova.conf')
        self.assertEquals(1, len(renderer._get_template.call_args_list))
        self.assertEquals(
            [call('nova.conf')], renderer._get_template.call_args_list)

    def test_render_template_by_munged_full_path_last(self):
        '''It loads a template by full path of config file second'''
        path = os.path.dirname(__file__)
        renderer = templating.OSConfigRenderer(templates_dir=path,
                                               openstack_release='folsom')
        tmp = MagicMock()
        tmp.render = MagicMock()
        e = TemplateNotFound('')
        renderer._get_template = MagicMock()
        renderer._get_template.side_effect = [e, tmp]
        renderer.register('/etc/nova/nova.conf', contexts=[])
        renderer.render('/etc/nova/nova.conf')
        self.assertEquals(2, len(renderer._get_template.call_args_list))
        self.assertEquals(
            [call('nova.conf'), call('etc_nova_nova.conf')],
            renderer._get_template.call_args_list)

    def test_render_template_by_basename(self):
        '''It renders template if it finds it by config file basename'''

    @patch(builtin_open)
    @patch.object(templating, 'get_loader')
    def test_write_out_config(self, loader, _open):
        '''It writes a templated config when provided a complete context'''
        self.context.set(interfaces=['fooservice'], context={'foo': 'bar'})
        self.renderer.register('/tmp/foo', [self.context])
        with patch.object(self.renderer, '_get_template') as _get_t:
            fake_tmpl = MockTemplate()
            _get_t.return_value = fake_tmpl
            self.renderer.write('/tmp/foo')
            _open.assert_called_with('/tmp/foo', 'wb')

    def test_write_all(self):
        '''It writes out all configuration files at once'''
        self.context.set(interfaces=['fooservice'], context={'foo': 'bar'})
        self.renderer.register('/tmp/foo', [self.context])
        self.renderer.register('/tmp/bar', [self.context])
        ex_calls = [
            call('/tmp/bar'),
            call('/tmp/foo'),
        ]
        with patch.object(self.renderer, 'write') as _write:
            self.renderer.write_all()
            self.assertEquals(sorted(ex_calls), sorted(_write.call_args_list))
            pass

    @patch.object(templating, 'get_loader')
    def test_reset_template_loader_for_new_os_release(self, loader):
        self.loader.set('')
        self.context.set(interfaces=['fooservice'], context={})
        loader.return_value = MockFSLoader('/tmp/foo')
        self.renderer.register('/tmp/foo', [self.context])
        self.renderer.render('/tmp/foo')
        loader.assert_called_with(os.path.dirname(__file__), 'folsom')
        self.renderer.set_release(openstack_release='grizzly')
        self.renderer.render('/tmp/foo')
        loader.assert_called_with(os.path.dirname(__file__), 'grizzly')

    @patch.object(templating, 'get_loader')
    def test_incomplete_context_not_reported_complete(self, loader):
        '''It does not recognize an incomplete context as a complete context'''
        self.context.set(interfaces=['fooservice'], context={})
        self.renderer.register('/tmp/foo', [self.context])
        self.assertNotIn('fooservice', self.renderer.complete_contexts())

    @patch.object(templating, 'get_loader')
    def test_complete_context_reported_complete(self, loader):
        '''It recognizes a complete context as a complete context'''
        self.context.set(interfaces=['fooservice'], context={'foo': 'bar'})
        self.renderer.register('/tmp/foo', [self.context])
        self.assertIn('fooservice', self.renderer.complete_contexts())

    @patch('os.path.isdir')
    def test_get_loader_no_templates_dir(self, isdir):
        '''Ensure getting loader fails with no template dir'''
        isdir.return_value = False
        self.assertRaises(templating.OSConfigException,
                          templating.get_loader,
                          templates_dir='/tmp/foo', os_release='foo')

    @patch('os.path.isdir')
    def test_get_loader_all_search_paths(self, isdir):
        '''Ensure loader reverse searches of all release template dirs'''
        isdir.return_value = True
        choice_loader = templating.get_loader('/tmp/foo',
                                              os_release='icehouse')
        dirs = [l.searchpath for l in choice_loader.loaders]

        common_tmplts = os.path.join(os.path.dirname(templating.__file__),
                                     'templates')
        expected = [['/tmp/foo/icehouse'],
                    ['/tmp/foo/havana'],
                    ['/tmp/foo/grizzly'],
                    ['/tmp/foo/folsom'],
                    ['/tmp/foo/essex'],
                    ['/tmp/foo/diablo'],
                    ['/tmp/foo'],
                    [common_tmplts]]
        self.assertEquals(dirs, expected)

    @patch('os.path.isdir')
    def test_get_loader_some_search_paths(self, isdir):
        '''Ensure loader reverse searches of some release template dirs'''
        isdir.return_value = True
        choice_loader = templating.get_loader('/tmp/foo', os_release='grizzly')
        dirs = [l.searchpath for l in choice_loader.loaders]

        common_tmplts = os.path.join(os.path.dirname(templating.__file__),
                                     'templates')

        expected = [['/tmp/foo/grizzly'],
                    ['/tmp/foo/folsom'],
                    ['/tmp/foo/essex'],
                    ['/tmp/foo/diablo'],
                    ['/tmp/foo'],
                    [common_tmplts]]
        self.assertEquals(dirs, expected)

    def test_register_template_with_list_of_contexts(self):
        '''Ensure registering a template with a list of context generators'''
        def _c1():
            pass

        def _c2():
            pass
        tmpl = templating.OSConfigTemplate(config_file='/tmp/foo',
                                           contexts=[_c1, _c2])
        self.assertEquals(tmpl.contexts, [_c1, _c2])

    def test_register_template_with_single_context(self):
        '''Ensure registering a template with a single non-list context'''
        def _c1():
            pass
        tmpl = templating.OSConfigTemplate(config_file='/tmp/foo',
                                           contexts=_c1)
        self.assertEquals(tmpl.contexts, [_c1])


class TemplatingStringTests(unittest.TestCase):
    def setUp(self):
        path = os.path.dirname(__file__)
        self.loader = FakeLoader()
        self.context = FakeContextGenerator()

        self.addCleanup(patch.object(templating,
                                     'apt_install').start().stop())
        self.addCleanup(patch.object(templating, 'log').start().stop())

        templating.FileSystemLoader = MockFSLoader
        templating.ChoiceLoader = MockChoiceLoader

        self.config_file = '/etc/confd/extensible.d/drop-in.conf'
        self.config_template = 'use: {{ fake_key }}'
        self.renderer = templating.OSConfigRenderer(templates_dir=path,
                                                    openstack_release='folsom')

    def test_render_template_from_string_full_context(self):
        '''
        Test rendering a specified config file with a string template
        and a context.
        '''

        context = {'fake_key': 'fake_val'}
        self.context.set(
            interfaces=['fooservice'],
            context=context
        )

        expected_output = 'use: {}'.format(context['fake_key'])

        self.renderer.register(
            config_file=self.config_file,
            contexts=[self.context],
            config_template=self.config_template
        )

        # should return a string given we render from an in-memory
        # template source
        output = self.renderer.render(self.config_file)

        self.assertEquals(output, expected_output)

    def test_render_template_from_string_incomplete_context(self):
        '''
        Test rendering a specified config file with a string template
        and a context.
        '''

        self.context.set(
            interfaces=['fooservice'],
            context={}
        )

        expected_output = 'use: '

        self.renderer.register(
            config_file=self.config_file,
            contexts=[self.context],
            config_template=self.config_template
        )

        # should return a string given we render from an in-memory
        # template source
        output = self.renderer.render(self.config_file)

        self.assertEquals(output, expected_output)

    def test_register_string_template_with_single_context(self):
        '''Template rendering from a provided string with a context'''
        def _c1():
            pass

        config_file = '/etc/confdir/custom-drop-in.conf'
        config_template = 'use: {{ key_available_in_c1 }}'
        tmpl = templating.OSConfigTemplate(
            config_file=config_file,
            contexts=_c1,
            config_template=config_template
        )

        self.assertEquals(tmpl.contexts, [_c1])
        self.assertEquals(tmpl.config_file, config_file)
        self.assertEquals(tmpl.config_template, config_template)
